from simses.commons.config.simulation.battery import BatteryConfig
from simses.commons.cycle_detection.cycle_detector import CycleDetector
from simses.technology.lithium_ion.cell.type import CellType
from simses.technology.lithium_ion.degradation.calendar.generic_cell import \
    GenericCellCalendarDegradationModel
from simses.technology.lithium_ion.degradation.cyclic.generic_cell import \
    GenericCellCyclicDegradationModel
from simses.technology.lithium_ion.degradation.degradation_model import DegradationModel


class GenericCellDegradationModel(DegradationModel):

    def __init__(self, cell_type: CellType, cycle_detector: CycleDetector, battery_config: BatteryConfig):
        super().__init__(cell_type, GenericCellCyclicDegradationModel(cell_type, cycle_detector, battery_config),
                         GenericCellCalendarDegradationModel(cell_type, battery_config), cycle_detector,
                         battery_config)


from abc import ABC, abstractmethod


class SolarIrradiationModel(ABC):
    """
        SolarIrradiationModel calculates the total incident solar irradiation on the selected housing object at any
        given time for a specified location.
        Solar irradiation time series data from various sources:
        """

    def __init__(self):
        super().__init__()
        pass

    @abstractmethod
    def get_heat_load(self, time: float) -> float:
        """
        This method is called from the system thermal model, and returns thermal power on container surfaces
        :param time:
        :return: thermal power, or 0 (depending on chosen sub-class)
        """
        pass

    @abstractmethod
    def get_global_horizontal_irradiance(self,time_step) -> float:
        """
        Returns the value of global horizontal irradiance for specified timestep for selected location (in W/m2)
        :param time_step:
        :return: irradiance, or 0 (depending on chosen sub-class)
        """
        pass

    @abstractmethod
    def create_instance(self):
        """
        reinstantiates the SolarIrradiationModel
        :return: SolarIrradiationModel
        """
        pass

    @abstractmethod
    def close(self):
        """
        closes all open resources in solar irradiation model

        Parameters
        ----------

        Returns
        -------

        """
        pass



import os
import pandas as pd

class Unificator :
    class bcolors :
        FAIL = '\033[1;91m'
        WARNING = '\033[1;93m'
        SUCCESS = '\033[1;32m'
        ENDC = '\033[1;37m'

    def __init__(self, print_df = True):
        self.print_df = print_df
        self.ROOTDIR = "../"
        self.paths = []
        self.semicolon_csvs = []
        self.coma_csvs = []
        self.comparisson = []
        self.comparisson_type = []
        self.sorting()
        self.updating_csvs(self.paths)

    def sorting(self):
        for subdir, dirs, files in os.walk(self.ROOTDIR):
            for file in files :
                path = os.path.join(subdir,file)
                if path[-4:]==".csv":
                    csv = open(path, "r")
                    last_line =csv.readlines()[-1]
                    try :
                        second_to_last_line = csv.readlines()[-2]
                    except :
                        second_to_last_line = ''
                    if ";" in last_line :
                        self.semicolon_csvs.append(path)
                        self.paths.append(path)

                    else :
                        # is it a one column Data frame with "," as decimal seperator
                        number_of_commas = last_line.count(",")
                        if number_of_commas > 1 or number_of_commas == 0 :
                            self.coma_csvs.append(path)
                            self.paths.append(path)

                        else:
                            self.semicolon_csvs.append(path)
                            self.paths.append(path)

    def updating_csvs(self,paths):
        for i in range(len(paths)):
            path = self.paths[i]
            name = path.rsplit("\\")[-1]
            if path in self.coma_csvs :
                print("\n {} : separator = \",\" only needs to update the headers \n".format(name))
                original = pd.read_csv(path,sep=",",engine="python")
                original.fillna("", inplace=True)
                if self.print_df:
                    print("The original dataframe : \n", original.head())
                original.columns = self.update_headers(original)
                if self.print_df:
                    print("The updated dataframe : \n",original.head())
                print(f"{self.bcolors.SUCCESS}Successfully updated " + name + f"{self.bcolors.ENDC}")
                original.to_csv(path, index=False)
            if path in self.semicolon_csvs :
                print("\n {} : separator = \";\" \n".format(name))
                check  = pd.read_csv(path,sep=';',engine="python")
                counter = 0
                sci_notation = 0
                for col in check.columns :
                    for x in check[col] :
                        if isinstance(x,str):
                            if ',' in x :
                                counter+=1
                            if 'e' in x :
                                sci_notation+=1

                        else :
                            pass
                print(counter)
                print(sci_notation)
                if counter >= 1 :
                    print('the decimal separator = \',\'')
                    original = pd.read_csv(path,sep=';',decimal=',',engine="python")
                    original.fillna("", inplace=True)
                    if self.print_df:
                        print("The updated dataframe : \n",original.head())
                    original.columns = self.update_headers(original)
                    updated = original.copy()
                    if sci_notation >= 1:
                        self.update_entries(updated)
                    if sci_notation == 0 :
                        pass
                    if self.print_df :
                        print("The updated dataframe : \n",updated.head())
                    self.compare(original_df=original,updated_df=updated, name=name)
                    # updated.to_csv(r'C:\Users\Lenovo\Desktop\simses save\Nouveau dossier/' + name, index=False)
                    updated.to_csv(path, index=False)

                if counter ==0 :
                    print('the decimal separator = \'.\'')
                    original = pd.read_csv(path,sep=';',decimal='.',engine="python",encoding="utf-8-sig")
                    original.fillna("", inplace=True)
                    if self.print_df:
                        print("The updated dataframe : \n",original.head())
                    original.columns = self.update_headers(original)
                    updated = original.copy()
                    self.update_entries(updated)
                    if self.print_df :
                        print("The updated dataframe : \n",updated.head())
                    self.compare(original_df=original,updated_df=updated, name=name)
                    updated.to_csv(path, index=False)




    def update_headers(self,df):
        header = []
        for col in df.columns :
            if "Unnamed:" in col :
                    header.append("")
            else :
                try :
                    header.append(col.replace(",","."))
                    if ";" in header[-1]:
                        header[-1] = header[-1].replace(";",",")
                    #stopping the mangle_dupe_cols,
                    #aka the automatic renaming of pandas of columns with the same name
                    if header[-1]==header[-2]+".1":
                        header[-1]=header[-2]
                except :
                    pass
        return header

    def update_entries(self, df):
        for col in df:
            try:
                df[col] = [x.replace(',', '.') for x in df[col]]
            except:
                pass

    def compare(self, original_df, updated_df,name):
        elements_original=[]
        elements_updated=[]
        original_types = original_df.copy()
        updated_types = updated_df.copy()

        for col in original_types.columns:
            original_types[col] = [type(x) for x in original_types[col]]
        for col in updated_types.columns:
            updated_types[col] = [type(x) for x in updated_types[col]]

        for col in original_df.columns:
            for x in original_df[col]:
                try :
                    elements_original.append(x.replace(",","."))
                except :
                    elements_original.append(x)

        for col in updated_df.columns:
            for x in updated_df[col]:
                elements_updated.append(x)

        if elements_original==elements_updated and original_types.equals(updated_types):
            print(f"{self.bcolors.SUCCESS}Successfully updated " + name + f"{self.bcolors.ENDC}")
        else:
            print(f"{self.bcolors.FAIL}Failed updating " + name + f"{self.bcolors.ENDC}")
            print(original_df.compare(updated_df))
            print(original_types.compare(updated_types))


instance = Unificator(print_df = False)


import csv
import os

import numpy

from simses.commons.config.data.power_electronics import PowerElectronicsConfig
from simses.commons.utils.utilities import format_float


def create_line_from(name: str, values: [float]) -> [str]:
    line: [str] = list()
    line.append(name)
    for value in values:
        line.append(format_float(value, decimals=4))
    return line


if __name__ == "__main__":
    max_power: float = 166e3
    config: PowerElectronicsConfig = PowerElectronicsConfig()
    voltage: float = 700
    power_range: [int] = range(-int(max_power), int(max_power), int(max_power / 200))
    file_name: str = os.path.dirname(__file__) + '/comparison.csv'

    converters: list = list()
    # converters.append(NottonAcDcConverter(max_power))
    # converters.append(AcDcConverterIdenticalStacked(3, 0.7, NottonAcDcConverter(max_power), config))
    # converters.append(AixControlAcDcConverter(max_power))
    # converters.append(M2bAcDcConverter(max_power))
    # converters.append(AcDcConverterIdenticalStacked(2, 0.7, NottonAcDcConverter(max_power), config))
    # converters.append(AcDcConverterIdenticalStacked(5, 0.7, NottonAcDcConverter(max_power), config))

    with open(file_name, 'w', newline='') as file:
        writer = csv.writer(file)
        writer.writerow(create_line_from('Power', numpy.array(power_range) / max_power))
        for converter in converters:
            efficiencies: [float] = list()
            for power in power_range:
                if power > 0:
                    power_dc: float = converter.to_dc(power, voltage)
                    try:
                        eff: float = power_dc / power
                    except ZeroDivisionError:
                        eff: float = 0.0
                    efficiencies.append(eff)
                else:
                    power_dc: float = converter.to_ac(power, voltage)
                    try:
                        eff: float = power / power_dc
                    except ZeroDivisionError:
                        eff: float = 0.0
                    efficiencies.append(eff)
            writer.writerow(create_line_from(type(converter).__name__, efficiencies))



from abc import ABC, abstractmethod


class ThermalController(ABC):
    """ This controller controls the temperature of the EL-stack by setting new values for the mass flow and the
    temperature of the water running through the stack. It is asumed that the water temperature coming out of the
    stack equals the stack temperature"""

    def __init__(self):
        super().__init__()

    @abstractmethod
    def calculate(self, stack_temperature: float, heat_stack: float, timestep: float, current_dens: float) -> None:
        pass

    @abstractmethod
    def get_heat_control_on(self) -> bool:
        pass

    @abstractmethod
    def get_h2o_flow(self) -> float:
        pass

    @abstractmethod
    def get_delta_water_temp_in(self) -> float:
        pass



from simses.commons.log import Logger
from simses.commons.state.technology.lithium_ion import LithiumIonState
from simses.technology.lithium_ion.cell.type import CellType
from simses.technology.lithium_ion.equivalent_circuit_model.equivalent_circuit import EquivalentCircuitModel


class RintModel(EquivalentCircuitModel):

    __ACCURACY: float = 1e-7

    def __init__(self, cell_type: CellType):
        super().__init__()
        self.__log: Logger = Logger(type(self).__name__)
        self.__cell_type: CellType = cell_type

    def update(self, time: float, battery_state: LithiumIonState) -> None:
        bs: LithiumIonState = battery_state
        self.__update_soc(time, bs)
        ocv: float = self.__cell_type.get_open_circuit_voltage(bs)  # V
        bs.voltage_open_circuit = ocv
        rint = self.__cell_type.get_internal_resistance(bs) * (1 + bs.resistance_increase)
        bs.internal_resistance = rint

    def __update_soc(self, time: float, bs: LithiumIonState) -> None:
        coulomb_eff = self.__cell_type.get_coulomb_efficiency(bs)
        self_discharge_rate: float = self.__cell_type.get_self_discharge_rate(bs) * (time - bs.time)
        # Ah (better As) counting
        bs.soe += bs.current * bs.voltage_open_circuit * (time - bs.time) / 3600.0 * coulomb_eff - self_discharge_rate
        bs.soc = bs.soe / bs.capacity
        # assuming rounding errors close to 0 or 1
        if abs(bs.soc) < self.__ACCURACY:
            self.__log.warn('SOC was tried to be set to a value of ' + str(bs.soc) + ' but adjusted to 0.0.')
            bs.soc = 0.0
        if abs(bs.soc - 1.0) < self.__ACCURACY:
            self.__log.warn('SOC was tried to be set to a value of ' + str(bs.soc) + ' but adjusted to 1.0.')
            bs.soc = 1.0

    def close(self) -> None:
        self.__log.close()



import json
import pickle
from configparser import ConfigParser


class ConfigGenerator:

    def __init__(self):
        self.__config: ConfigParser = ConfigParser()

    def load_config_from(self, file: str) -> None:
        """
        Loads config from given file

        Parameters
        ----------
        file :
            path to config file

        Returns
        -------

        """
        self.__config.read(file)

    def get_config(self) -> ConfigParser:
        """

        Returns
        -------
        ConfigParser:
            copy of config setup to be passed to SimSES
        """
        rep = pickle.dumps(self.__config)
        new_config = pickle.loads(rep)
        return new_config

    def show(self) -> None:
        """
        Printing config setup

        Returns
        -------

        """
        config: dict = {section: dict(self.__config[section]) for section in self.__config.sections()}
        print(json.dumps(config, sort_keys=True, indent=4))

    def _set(self, section: str, option: str, value: str) -> None:
        if value is None:
            return
        config: ConfigParser = self.__config
        if section not in config.sections():
            config.add_section(section)
        config.set(section, option, value)

    def _set_bool(self, section: str, option: str, value: bool) -> None:
        self._set(section, option, 'True' if value else 'False')

    def _clear(self, section: str, option: str) -> None:
        self._set(section, option, '')

    def _get(self, section: str, option: str) -> str:
        try:
            return self.__config[section][option]
        except KeyError as err:
            raise err

    def _add(self, section: str, option: str, value: str) -> None:
        prop: str = self._get(section, option)
        prop += '\n' + value
        self._set(section, option, prop)

    def _get_id_from(self, section: str, option: str, delimiter: str = '\n') -> str:
        try:
            prop: str = self._get(section, option)
            return '_' + str(len(prop.split(delimiter)))
        except KeyError:
            return '_' + str(0)



from simses.commons.state.technology.storage import StorageTechnologyState


class RedoxFlowState(StorageTechnologyState):
    """
    Current physical state of the redox_flow system with the main electrical parameters.
    """

    SYSTEM_AC_ID = 'StorageSystemAC'
    SYSTEM_DC_ID = 'StorageSystemDC'
    SOC = 'SOC in p.u.'
    SOC_STACK = 'SOC in stack in p.u.'
    SOH = 'State of health in p.u.'
    VOLTAGE = 'Voltage in V'
    OPEN_CIRCUIT_VOLTAGE = 'Open circuit voltage (OCV) in V'
    CURRENT = 'Current in A'
    CAPACITY = 'Capacity in Wh'
    INTERNAL_RESISTANCE = 'Internal resistance in Ohm'
    POWER_LOSS = 'Power loss in W'
    POWER_IN = 'Power input in W'
    POWER = 'Power output in W'
    PUMP_POWER = 'Pump power in W'
    PRESSURE_LOSS_ANOLYTE = 'Pressure loss anolyte in W'
    PRESSURE_LOSS_CATHOLYTE = 'Pressure loss catholyte in W'
    FLOW_RATE_ANOLYTE = 'Anolyte flow rate in m^3/s'
    FLOW_RATE_CATHOLYTE = 'Catholyte flow rate in m^3/s'
    PRESSURE_DROP_ANOLYTE = 'Anolyte pressure drop in Pa'
    PRESSURE_DROP_CATHOLYTE = 'Catholyte pressure drop in Pa'
    FULFILLMENT = 'Fulfillment in p.u.'
    TEMPERATURE = 'Electrolyte and stack temperature in K'
    TIME_PUMP = 'Time of pump on or off in s'
    MAX_CHARGE_POWER = 'Maximum charging power in W'
    MAX_DISCHARGE_POWER = 'Maximum discharging power in W'

    def __init__(self, system_id: int, storage_id: int):
        super().__init__()
        self._initialize()
        self.set(self.SYSTEM_AC_ID, system_id)
        self.set(self.SYSTEM_DC_ID, storage_id)

    @property
    def soc(self) -> float:
        return self.get(self.SOC)

    @soc.setter
    def soc(self, value: float) -> None:
        self.set(self.SOC, value)

    @property
    def soc_stack(self) -> float:
        return self.get(self.SOC_STACK)

    @soc_stack.setter
    def soc_stack(self, value: float) -> None:
        self.set(self.SOC_STACK, value)

    @property
    def soh(self) -> float:
        return self.get(self.SOH)

    @soh.setter
    def soh(self, value: float) -> None:
        self.set(self.SOH, value)

    @property
    def voltage(self) -> float:
        return self.get(self.VOLTAGE)

    @voltage.setter
    def voltage(self, value: float) -> None:
        self.set(self.VOLTAGE, value)

    @property
    def open_circuit_voltage(self) -> float:
        return self.get(self.OPEN_CIRCUIT_VOLTAGE)

    @open_circuit_voltage.setter
    def open_circuit_voltage(self, value: float) -> None:
        self.set(self.OPEN_CIRCUIT_VOLTAGE, value)

    @property
    def current(self) -> float:
        return self.get(self.CURRENT)

    @current.setter
    def current(self, value: float) -> None:
        self.set(self.CURRENT, value)

    @property
    def capacity(self) -> float:
        return self.get(self.CAPACITY)

    @capacity.setter
    def capacity(self, value: float) -> None:
        self.set(self.CAPACITY, value)

    @property
    def internal_resistance(self) -> float:
        return self.get(self.INTERNAL_RESISTANCE)

    @internal_resistance.setter
    def internal_resistance(self, value: float) -> None:
        self.set(self.INTERNAL_RESISTANCE, value)

    @property
    def power_loss(self) -> float:
        return self.get(self.POWER_LOSS)

    @power_loss.setter
    def power_loss(self, value: float) -> None:
        self.set(self.POWER_LOSS, value)

    @property
    def power_in(self) -> float:
        return self.get(self.POWER_IN)

    @power_in.setter
    def power_in(self, value: float) -> None:
        self.set(self.POWER_IN, value)

    @property
    def power(self) -> float:
        return self.get(self.POWER)

    @power.setter
    def power(self, value: float) -> None:
        self.set(self.POWER, value)

    @property
    def pump_power(self) -> float:
        return self.get(self.PUMP_POWER)

    @pump_power.setter
    def pump_power(self, value: float) -> None:
        self.set(self.PUMP_POWER, value)

    @property
    def pressure_loss_anolyte(self) -> float:
        return self.get(self.PRESSURE_LOSS_ANOLYTE)

    @pressure_loss_anolyte.setter
    def pressure_loss_anolyte(self, value: float) -> None:
        self.set(self.PRESSURE_LOSS_ANOLYTE, value)

    @property
    def pressure_loss_catholyte(self) -> float:
        return self.get(self.PRESSURE_LOSS_CATHOLYTE)

    @pressure_loss_catholyte.setter
    def pressure_loss_catholyte(self, value: float) -> None:
        self.set(self.PRESSURE_LOSS_CATHOLYTE, value)

    @property
    def flow_rate_anolyte(self) -> float:
        return self.get(self.FLOW_RATE_ANOLYTE)

    @flow_rate_anolyte.setter
    def flow_rate_anolyte(self, value: float) -> None:
        self.set(self.FLOW_RATE_ANOLYTE, value)

    @property
    def flow_rate_catholyte(self) -> float:
        return self.get(self.FLOW_RATE_CATHOLYTE)

    @flow_rate_catholyte.setter
    def flow_rate_catholyte(self, value: float) -> None:
        self.set(self.FLOW_RATE_CATHOLYTE, value)

    @property
    def pressure_drop_anolyte(self) -> float:
        return self.get(self.PRESSURE_DROP_ANOLYTE)

    @pressure_drop_anolyte.setter
    def pressure_drop_anolyte(self, value: float) -> None:
        self.set(self.PRESSURE_DROP_ANOLYTE, value)

    @property
    def pressure_drop_catholyte(self) -> float:
        return self.get(self.PRESSURE_DROP_CATHOLYTE)

    @pressure_drop_catholyte.setter
    def pressure_drop_catholyte(self, value: float) -> None:
        self.set(self.PRESSURE_DROP_CATHOLYTE, value)

    @property
    def fulfillment(self) -> float:
        return self.get(self.FULFILLMENT)

    @fulfillment.setter
    def fulfillment(self, value: float) -> None:
        self.set(self.FULFILLMENT, value)

    @property
    def temperature(self) -> float:
        return self.get(self.TEMPERATURE)

    @temperature.setter
    def temperature(self, value: float) -> None:
        self.set(self.TEMPERATURE, value)

    @property
    def time_pump(self) -> float:
        return self.get(self.TIME_PUMP)

    @time_pump.setter
    def time_pump(self, value: float) -> None:
        self.set(self.TIME_PUMP, value)

    @property
    def id(self) -> str:
        return 'REDOXFLOW' + str(self.get(self.SYSTEM_AC_ID)) + str(self.get(self.SYSTEM_DC_ID))

    @property
    def is_charge(self) -> bool:
        return self.power > 0

    @classmethod
    def sum_parallel(cls, states: []):
        pass

    @classmethod
    def sum_serial(cls, states: []):
        pass

    @property
    def max_charge_power(self) -> float:
        return self.get(self.MAX_CHARGE_POWER)

    @max_charge_power.setter
    def max_charge_power(self, value: float) -> None:
        self.set(self.MAX_CHARGE_POWER, value)

    @property
    def max_discharge_power(self) -> float:
        return self.get(self.MAX_DISCHARGE_POWER)

    @max_discharge_power.setter
    def max_discharge_power(self, value: float) -> None:
        self.set(self.MAX_DISCHARGE_POWER, value)



from datetime import datetime

import numpy
import pytz

from simses.analysis.data.energy_management import EnergyManagementData
from simses.analysis.data.system import SystemData
from simses.analysis.evaluation.economic.revenue_stream.abstract_revenue_stream import RevenueStream
from simses.analysis.evaluation.result import EvaluationResult, Unit, Description
from simses.commons.config.analysis.economic import EconomicAnalysisConfig
from simses.commons.log import Logger
from simses.commons.utils.utilities import add_month_to, add_year_to, get_average_from, get_maximum_from


class DemandChargeReduction(RevenueStream):

    class BillingPeriod:
        MONTHLY: str = 'monthly'
        YEARLY: str = 'yearly'
        OPTIONS: [str] = [MONTHLY, YEARLY]

    def __init__(self, energy_management_data: EnergyManagementData, system_data: SystemData,
                 economic_analysis_config: EconomicAnalysisConfig):
        super().__init__(energy_management_data, system_data, economic_analysis_config)
        self.__log: Logger = Logger(type(self).__name__)
        self.__billing_period = economic_analysis_config.demand_charge_billing_period
        self.__demand_charge = economic_analysis_config.demand_charge_price
        self.__demand_charge_average_interval = economic_analysis_config.demand_charge_average_interval
        self.__key_results: [EvaluationResult] = list()
        self.__seconds_per_year = 24 * 60 * 60 * 365
        if self.__billing_period not in self.BillingPeriod.OPTIONS:
            raise Exception('Please configure demand charge billing period to one of the following options: ' +
                            str(self.BillingPeriod.OPTIONS) + '. The current billing cycle is set to ' +
                            str(self.__billing_period) + '.')

    def __get_demand_charge_in_billing_period_for(self, time: numpy.ndarray, power: numpy.ndarray) \
            -> ([float], [datetime]):
        """
         Calculate demand charge based on absolute power values in each billing periods and detemrine billing periods.
         Here, demand charges are not scaled yet:
         If only one day is in the billing period, the full demand charge will apply

        Parameters
        ----------
        time :
            series of unix time stamps as numpy array
        power:
            series of site power values as numpy array

        Returns
        -------
        list of float
            List of float, representing demand charges in each billing period
        list of datetime
            List of datetime, representing beginning and end of each billing period
        """
        power_time_series: numpy.ndarray = numpy.column_stack([time, abs(power)])
        billing_periods: list = list()
        # determine start and end of first billing period in unix timestamps
        billing_period_start: float = self.__get_initial_billing_period_for(time[0]).timestamp()
        billing_periods.append(datetime.fromtimestamp(billing_period_start, tz=pytz.UTC))
        billing_period_end: float = self.__get_next_billing_period(billing_period_start).timestamp()

        # determine interval end of first interval for averging power values (e.g. 15min in Germany)
        interval_end = time[0] - time[0] % self.__demand_charge_average_interval + self.__demand_charge_average_interval

        # initialise lists
        power_interval: list = list()
        power_avg_in_interval: list = list()
        power_max_for_billing_period: list = list()
        for tstmp, value in power_time_series:
            power_interval.append(value)
            if tstmp >= interval_end:  # get average power values for demand charge interval
                power_avg_in_interval.append(get_average_from(power_interval))
                power_interval.clear()
                interval_end += self.__demand_charge_average_interval
            if tstmp >= billing_period_end:  # get maximum power values for billing period
                # scale values accordingly to days that have passed in the respective billing period
                power_max_for_billing_period.append(get_maximum_from(power_avg_in_interval))
                power_avg_in_interval.clear()
                billing_periods.append(datetime.fromtimestamp(billing_period_end, tz=pytz.UTC))
                billing_period_start = billing_period_end
                billing_period_end = self.__get_next_billing_period(billing_period_start).timestamp()
        # add remaining values
        power_avg_in_interval.append(get_average_from(power_interval))
        power_max_for_billing_period.append(get_maximum_from(power_avg_in_interval))
        # add end of last billing period
        billing_periods.append(datetime.fromtimestamp(billing_period_end, tz=pytz.UTC))
        # convert maximum power values into demand charge
        return list(numpy.array(power_max_for_billing_period) * self.__demand_charge / 1000.0), billing_periods

    @staticmethod
    def __get_cashflow_time_series(time: [float], demand_charges: [float], billing_periods: [datetime]) -> [float]:
        """
         Turns list of billing periods and demand charges for each billing period into a
         list of cashflow values for each timestep.

        Parameters
        ----------
        time :
            series of unix time stamps as list of float
        demand_charges:
            series of demand_charges as list of float
        billing_periods:
            series of billing periods as list of datetime

        Returns
        -------
        list of float
            List of float, representing demand charges for each timestep
        """
        cashflow_list = list()
        i = 0  # counter for current billing period
        delta_t = time[1] - time[0]
        start_period = billing_periods[i].timestamp()
        end_period = billing_periods[i+1].timestamp()
        for t in time:
            if t > end_period:
                i += 1
                start_period = billing_periods[i].timestamp()
                end_period = billing_periods[i + 1].timestamp()
            cashflow = delta_t / (end_period - start_period) * demand_charges[i]
            cashflow_list.append(cashflow)
        return cashflow_list

    def __get_initial_billing_period_for(self, tstmp: float) -> datetime:
        date = datetime.fromtimestamp(tstmp, tz=pytz.UTC)
        # start of month
        date = date.replace(day=1, hour=0, minute=0, second=0, microsecond=0)
        if self.__billing_period == self.BillingPeriod.YEARLY:
            # start of year
            date = date.replace(month=1)
        return date

    def __get_next_billing_period(self, tstmp: float) -> datetime:
        date = datetime.fromtimestamp(tstmp, tz=pytz.UTC)
        if self.__billing_period == self.BillingPeriod.MONTHLY:
            date = add_month_to(date)
        else:  # self.__billing_period == self.BillingPeriod.YEARLY
            date = add_year_to(date)
        return date

    def __print_warning(self):
        time = self._energy_management_data.time
        if time[1] - time[0] > self.__demand_charge_average_interval:
            self.__log.warn('Simulation timestep is larger than DEMAND_CHARGE_AVERAGE_INTERVAL (' +
                             str(self.__demand_charge_average_interval) + ' sec). Economic evaluation for peak shaving '
                                                                          'may not be representative.')

    def get_cashflow(self) -> numpy.ndarray:
        self.__print_warning()
        # get data for calculation
        time: numpy.ndarray = self._energy_management_data.time
        load_power: numpy.ndarray = self._energy_management_data.load_power
        pv_power: numpy.ndarray = self._energy_management_data.pv_power
        storage_power: numpy.ndarray = self._system_data.power
        # calculate power arrays
        grid_power_base: numpy.ndarray = load_power - pv_power
        grid_power_storage: numpy.ndarray = load_power - pv_power + storage_power

        # calculate demand charge in each billing period
        demand_charge_base, billing_periods = self.__get_demand_charge_in_billing_period_for(time, grid_power_base)
        demand_charge_storage, billing_periods = self.__get_demand_charge_in_billing_period_for(time, grid_power_storage)

        # turn into time series
        demand_charge_base_ts = self.__get_cashflow_time_series(list(time), demand_charge_base, billing_periods)
        demand_charge_storage_ts = self.__get_cashflow_time_series(list(time), demand_charge_storage, billing_periods)

        # calculate in project years
        demand_charge_per_year_base = self.cash_time_series_to_project_years(demand_charge_base_ts, time)
        demand_charge_per_year_storage = self.cash_time_series_to_project_years(demand_charge_storage_ts, time)

        # append demand charges to key results
        self.__key_results.append(EvaluationResult(Description.Economical.DemandCharges.COST_WITHOUT_STORAGE,
                                                   Unit.EURO, demand_charge_per_year_base))
        self.__key_results.append(EvaluationResult(Description.Economical.DemandCharges.COST_WITH_STORAGE,
                                                   Unit.EURO, demand_charge_per_year_storage))
        # self.__key_results.append(
        #     EvaluationResult('Demand charges each billing period without storage', Unit.EURO, demand_charge_base))
        #  self.__key_results.append(
        #     EvaluationResult('Demand charges each billing period with storage', Unit.EURO, demand_charge_storage))
        revenue_nondiscounted = numpy.array(demand_charge_per_year_base) - numpy.array(demand_charge_per_year_storage)
        return revenue_nondiscounted

    def get_evaluation_results(self):
        return self.__key_results

    def get_assumptions(self):
        assumptions: [EvaluationResult] = list()
        assumptions.append(EvaluationResult(Description.Economical.DemandCharges.CYCLE, Unit.NONE,
                                            self.__billing_period))
        assumptions.append(EvaluationResult(Description.Economical.DemandCharges.PRICE, Unit.EURO_PER_KW,
                                            self.__demand_charge))
        assumptions.append(EvaluationResult(Description.Economical.DemandCharges.INTERVAL, Unit.MINUTES,
                                            self.__demand_charge_average_interval / 60.0))
        return assumptions

    def close(self):
        self.__log.close()



import datetime

from simses.commons.config.simulation.general import GeneralSimulationConfig
from simses.commons.config.simulation.profile import ProfileConfig
from simses.commons.profile.technical.soc import SocProfile
from simses.commons.state.energy_management import EnergyManagementState
from simses.commons.state.system import SystemState
from simses.logic.energy_management.strategy.operation_priority import OperationPriority
from simses.logic.energy_management.strategy.operation_strategy import OperationStrategy
from simses.commons.profile.technical.binary import BinaryProfile
from simses.commons.config.simulation.energy_management import EnergyManagementConfig
import numpy as np
from copy import copy
import time as time_measurement

class ElectricVehicleSOC(OperationStrategy):

    """
    ElectricVehicleSOC is a basic operation strategy which simulates an EV during trips and recharge. The algorithm
    requires the following profiles and parameters:
    - An SOC profile that contains the SOC of the vehicle. The charging of the vehicle can be changed but also the
    original charging SOC-change can be used
    - A binary profile that contains ones when the EV is parked at home and zeros if not
    - A string for the parameter 'EV_CHARGING_STRATEGY' which can be:
        - "Original": Uses the original SOC-values for the charging
        - "Uncontrolled": Always recharge, if SOC<100% and parked at "home"
        - "Mean_Power": This strategy checks the next departure time (perfect foresight) and chooses the charging power
            to reach 100% SOC exactly at departure time.
        - "Paused, integer threshold value": This strategy charges immediately after arrival until a threshold value
            (e.g. 0.8 -> 80% SOC). Then the charging is paused. The charging continues with maximal grid power at the
            time (t = Delta_E/P) that allows 100% SOC to be reached at the next departure time (with 30 minutes buffer).
    - The maximal grid power is set in the 'ENERGY_MANAGEMENT' section as 'MAX_POWER'

    Whenever the EV is not at home, the EMS just follows the load profile. If the car is parked at home, it is
    recharged depending on 'EV_CHARGING_STRATEGY'. The algorithm can also be used for electric buses.
    """

    def __init__(self, config: GeneralSimulationConfig, profile_config: ProfileConfig, ems_config: EnergyManagementConfig):
        super().__init__(OperationPriority.MEDIUM)

        self.__soc_profile = SocProfile(config, profile_config) # SOC profile is used in this strategy
        self.__binary_profile = BinaryProfile(config, profile_config)
        self.__soc: float = 0.0
        self.__binary: bool = False
        self.__charging_strategy = ems_config.ev_charging_strategy
        self.__max_ac_grid_power = ems_config.max_power  # Here maximal grid power

        self.__timestep = config.timestep
        self.__Wh_to_Ws = 3600
        self.__next_time_zero = None
        self.__soc_at_nextzero = None
        self.__target_soc = None

        self.__config = config
        self.__profile_config = profile_config
        self.__ems_config = ems_config

        self.__helper_class = HelperClassForProfileCopies(config, profile_config, ems_config)
        if self.__charging_strategy[0:20] == 'Paused_to_target_SOC':
            self.__threshold = float(self.__charging_strategy[-3:])


    def next(self, time: float, system_state: SystemState, power: float = 0) -> float:

        self.__soc = self.__soc_profile.next(time) # SOC
        if self.__soc < 0:
            print("Error: Desired SOC < 0")

        temp_binary = np.ceil(self.__binary_profile.next(time))
        # rounds binary upwards. If interpolation/no data available and next existing value is 1 all values inbetween
        # are denoted as 1. Thus, charging is possible.
        if temp_binary!=0 and temp_binary!=1:
            raise ValueError("The binary profile contains values different from 0 or 1!")
        self.__binary = bool(temp_binary)

        if not self.__binary: # EV is driving

            soc_dif = self.__soc - system_state.soc
            capacity_dif = soc_dif * system_state.capacity * self.__Wh_to_Ws
            power = capacity_dif / self.__timestep
            # E-Bus profile contained weired changes of SOC (Delta >0) while driving (not recuperation!?)
            # which is why a student differentiated here: if soc_dif > 0 => power=0

        else: # EV is parked
            if np.round(system_state.soc,5)==1: # if already fully charged
                power = 0
            elif self.__charging_strategy == 'Original': # just follow SOC
                soc_dif = self.__soc - system_state.soc
                capacity_dif = soc_dif * system_state.capacity * self.__Wh_to_Ws
                power = capacity_dif / self.__timestep

            elif self.__charging_strategy == 'Uncontrolled_to_target_SOC': # Charge with full power to reach a target SOC

                if self.__next_time_zero is None:
                    self.__next_time_zero = self.__helper_class.binary_change_function(time, self.__binary)
                    next_time_zero = datetime.datetime.fromtimestamp(self.__next_time_zero)

                    if self.__next_time_zero >= time + self.__timestep:
                        self.__target_soc = self.__helper_class.specific_time_soc(time, self.__next_time_zero)
                    else:
                        self.__target_soc = self.__soc

                if system_state.soc < self.__target_soc:
                    power = self.__max_ac_grid_power    # charge with max power
                else:
                    power = 0

                if self.__next_time_zero is not None:
                    if self.__next_time_zero - time <= self.__timestep:
                        self.__next_time_zero = None


            elif self.__charging_strategy == 'Mean_to_target_SOC': # Charge with mean power to be at target soc at departure time

                if self.__next_time_zero is None:
                    self.__next_time_zero = self.__helper_class.binary_change_function(time, self.__binary)
                    if self.__next_time_zero >= time + self.__timestep:
                        self.__target_soc = self.__helper_class.specific_time_soc(time, self.__next_time_zero)
                    else:
                        self.__target_soc = self.__soc

                charging_duration = self.__next_time_zero - time
                capacity_to_charge = (self.__target_soc - system_state.soc) * system_state.capacity * self.__Wh_to_Ws

                if charging_duration > 0:
                    power = capacity_to_charge / charging_duration
                else:
                    power = self.__max_ac_grid_power
                    self.__next_time_zero = None

                if self.__next_time_zero is not None:
                    if self.__next_time_zero - time <= self.__timestep:
                        self.__next_time_zero = None

            elif self.__charging_strategy[0:20] == 'Paused_to_target_SOC': # Charge with mean power to be at target soc at departure time

                if self.__threshold > 0.9: # Threshold to charge to should not be bigger than target SOC
                    if self.__next_time_zero is None:
                        self.__next_time_zero = self.__helper_class.binary_change_function(time, self.__binary)
                    if self.__next_time_zero >= time + self.__timestep:
                        self.__target_soc = self.__helper_class.specific_time_soc(time, self.__next_time_zero)
                    else:
                        self.__target_soc = self.__soc
                    current_threshold = self.__target_soc
                else:
                    current_threshold = self.__threshold

                if system_state.soc < current_threshold:
                    power = self.__max_ac_grid_power
                else:
                    if self.__next_time_zero is None:
                        self.__next_time_zero = self.__helper_class.binary_change_function(time, self.__binary)
                        if self.__next_time_zero >= time + self.__timestep:
                            self.__target_soc = self.__helper_class.specific_time_soc(time, self.__next_time_zero)
                        else:
                            self.__target_soc = self.__soc
                    else:
                        charging_duration = (self.__target_soc - system_state.soc)*system_state.capacity*self.__Wh_to_Ws/self.__max_ac_grid_power
                        charging_restart_time = self.__next_time_zero - charging_duration - 1800
                        # start recharging after the pause, buffer of 30 minutes (1800s)
                        if time >= charging_restart_time and system_state.soc < self.__target_soc:
                            power = self.__max_ac_grid_power
                        else:
                            power = 0

                if self.__next_time_zero is not None:
                    if self.__next_time_zero - time <= self.__timestep:
                        self.__next_time_zero = None


            elif self.__charging_strategy == 'Uncontrolled': # Recharge with max AC power when parked
                power = self.__max_ac_grid_power

            elif self.__charging_strategy == 'Mean_power': # Charge with mean power to be fully charged at departure time

                if self.__next_time_zero is None:
                    self.__next_time_zero = self.__binary_profile.next_change_in_binary(time, self.__binary)

                charging_duration = self.__next_time_zero - time
                capacity_to_charge = (1 - system_state.soc) * system_state.capacity * self.__Wh_to_Ws
                if charging_duration > 0:
                    power = capacity_to_charge / charging_duration
                else:
                    power = self.__max_ac_grid_power

                if charging_duration == self.__timestep:
                    self.__next_time_zero = None

            elif self.__charging_strategy[0:6] == 'Paused':
                threshold = float(self.__charging_strategy[-3:])
                if system_state.soc < threshold:
                    power = self.__max_ac_grid_power
                else:
                    if self.__next_time_zero is None:
                        self.__next_time_zero = self.__binary_profile.next_change_in_binary(time, self.__binary)
                    else:
                        charging_duration = (1-threshold)*system_state.capacity*self.__Wh_to_Ws/self.__max_ac_grid_power
                        charging_restart_time = self.__next_time_zero - charging_duration - 1800
                        # start recharging after the pause, buffer of 30 minutes (1800s)
                        if time < charging_restart_time:
                            power = 0
                        else:
                            power = self.__max_ac_grid_power
                            if self.__next_time_zero - time < self.__timestep:
                                self.__next_time_zero = None
        return power


    def update(self, energy_management_state: EnergyManagementState) -> None:
        energy_management_state.binary = int(self.__binary)

    def clear(self) -> None:
        pass

    def close(self) -> None:
        self.__soc_profile.close()
        self.__binary_profile.close()


class HelperClassForProfileCopies:

    def __init__(self, config: GeneralSimulationConfig, profile_config: ProfileConfig, ems_config: EnergyManagementConfig):

        # open second soc and binary profile for forecasts
        # This is necessary, as next function of profiles of instance of ElectricVehicleSOC-class saves previous values
        # Thus, those profiles cannot be used for forecasting SOC or binary values
        self.__soc_profile_copy = SocProfile(config, profile_config)
        self.__binary_profile_copy = BinaryProfile(config, profile_config)
        self.__timestep = config.timestep

    def binary_change_function(self, time, current_binary):
        # This function returns the time when the binary profile changes its value
        # from 0 to 1 or 1 to 0

        return self.__binary_profile_copy.next_change_in_binary(time, current_binary)

    def specific_time_soc(self, current_time, target_time) -> float:
        # This function returns the soc value of a specific point in time

        for time_counter in range(int(current_time), int(target_time),int(self.__timestep)):
            soc = self.__soc_profile_copy.next(time_counter)

        if 'soc' not in locals():
            soc = 1
            print('Error in specific_time_soc of HelperClassForProfileCopies of ElectricVehicleSOC. '
                  'Is target time > current time?')

        return soc





import numpy

from simses.commons.constants import Hydrogen
from simses.commons.state.technology.electrolyzer import ElectrolyzerState
from simses.technology.hydrogen.electrolyzer.stack.pem_analytic.pressure_model import \
    PemPressureModel


class PemMembraneModel:

    # TODO Earnings per share?
    EPS = numpy.finfo(float).eps
    GEOM_AREA_CELL = 1225  # cm2  example from Siemens from: PEM-Elektrolyse-Systeme zur Anwendung in Power-to-Gas Anlagen
    DIFF_COEF_H2 = 4.65 * 10 ** (-11)  # mol/(cm s bar)
    DIFF_COEF_O2 = 2 * 10 ** (-11)  # mol/(cm s bar)
    DP_COEF_H2 = 2 * 10 ** (-11)  # mol/(cm s bar)
    WATER_DRAG_COEFF = 0.27  # molH20/ molH+

    def __init__(self, pressure_model: PemPressureModel, geom_area_cell: float = GEOM_AREA_CELL, thickness: float = 200 * 10 ** (-4),
                 humidification: float = 25.0):
        self.__thickness: float = thickness  # cm
        self.__LAMBDA: float = humidification  # degree of humidification
        self.__geom_area_cell: float = geom_area_cell  # cm2
        self.__pressure_model: PemPressureModel = pressure_model

    def get_geometric_area_cell(self) -> float:
        return self.__geom_area_cell

    def resistance(self, stack_temperature) -> float:
        conductivity_nafion = (0.005139 * self.__LAMBDA - 0.00326) * \
                              numpy.exp(1268 * (1 / (303) - 1 / (stack_temperature + 273.15)))  # S/cm
        return self.__thickness / conductivity_nafion

    def get_h2_permeation_for_cell(self, state: ElectrolyzerState, current_cell: float) -> float:
        """
        permeation of hydrogen through membrane due to diffusion because of differential partial pressures
        Parameters
        ----------
        part_pressure_h2 :
        part_pressure_o2 :

        Returns
        -------
        float:
            mol/s
        """
        current_dens_cell = current_cell / self.__geom_area_cell
        part_pressure_h2: float = self.__pressure_model.get_partial_pressure_h2(state, current_dens_cell)
        sat_pressure_h2o: float = self.__pressure_model.get_sat_pressure_h2o(state.temperature)
        part_pressure_o2: float = self.__pressure_model.get_partial_pressure_o2(state, current_dens_cell)
        if part_pressure_h2 + sat_pressure_h2o <= 1 + self.EPS:  # prevention of negatative pressures at cathode side
            return 0
        return self.DIFF_COEF_H2 * part_pressure_h2 / self.__thickness * self.__geom_area_cell + self.DP_COEF_H2 * \
               (part_pressure_h2 - part_pressure_o2) / self.__thickness * self.__geom_area_cell  # mol/s

    def get_o2_permeation_for_cell(self, state: ElectrolyzerState, current_cell: float) -> float:
        """
        permeation of oxygen through membrane due to diffusion because of differential partial pressures

        Parameters
        ----------
        part_pressure_o2 :

        Returns
        -------

        """
        current_dens_cell = current_cell / self.__geom_area_cell
        part_pressure_h2: float = self.__pressure_model.get_partial_pressure_h2(state, current_dens_cell)
        part_pressure_o2: float = self.__pressure_model.get_partial_pressure_o2(state, current_dens_cell)
        if part_pressure_h2 == 0 or current_cell == 0:  # prevention of negative pressures at anode side
            return 0
        return self.DIFF_COEF_O2 * part_pressure_o2 / self.__thickness * self.__geom_area_cell  # mol/s

    def get_h2o_permeation_for_cell(self, current_cell: float) -> float:
        return self.WATER_DRAG_COEFF * current_cell / Hydrogen.FARADAY_CONST



from abc import ABC, abstractmethod
from simses.commons.log import Logger
from simses.system.housing.layer import Layer


class Housing(ABC):
    """ class to specify the housing of the storage system"""

    def __init__(self, layer_inner: Layer = None, layer_mid: Layer = None, layer_outer: Layer = None, unavailable_volume: float = 0, default_scale: int=0):
        super().__init__()
        self.__log: Logger = Logger(type(self).__name__)
        self.__layer_inner = layer_inner
        self.__layer_mid = layer_mid
        self.__layer_outer = layer_outer
        self.__internal_component_volume: float = 0.0
        self.__scale_factor: int = default_scale
        # self.__volume_usability_factor: float = 0.2
        self.__volume_usability_factor: float = 1 - unavailable_volume / self.internal_volume

    def add_component_volume(self, volume: float):
        self.__internal_component_volume += volume
        while self.__internal_component_volume > self.__volume_usability_factor * self.internal_volume:
            self.__scale_factor += 1
            self.__layer_inner.update_scale()
            self.__layer_mid.update_scale()
            self.__layer_outer.update_scale()
            self.__log.info('Increasing number of containers to ' + str(self.__scale_factor))

    @property
    def inner_layer(self) -> Layer:
        """
        Access to layer attributes of inner-layer such as temperature, dimensions, mass, thermal properties

        Returns
        -------

        """
        return self.__layer_inner

    @property
    def mid_layer(self) -> Layer:
        """

        Access to layer attributes of mid-layer such as temperature, dimensions, mass, thermal properties

        Returns
        -------

        """
        return self.__layer_mid

    @property
    def outer_layer(self) -> Layer:
        """
        Access to layer attributes of outer-layer such as temperature, dimensions, mass, thermal properties

        Returns
        -------

        """
        return self.__layer_outer

    @property
    def internal_volume(self) -> float:
        """
        Returns internal volume of housing in m3

        Returns
        -------

        """
        return (self.inner_layer.length * self.inner_layer.breadth * self.inner_layer.height) * self.__scale_factor

    @property
    def internal_air_volume(self) -> float:
        """
        Returns internal air volume of housing in m3

        Returns
        -------

        """
        return self.internal_volume - self.__internal_component_volume

    @property
    @abstractmethod
    def azimuth(self) -> float:
        pass

    @property
    @abstractmethod
    def albedo(self) -> float:
        pass

    def get_number_of_containers(self) -> int:
        return self.__scale_factor

    def close(self) -> None:
        self.__log.close()



import numpy as np

from simses.commons.log import Logger
from simses.commons.state.system import SystemState
from simses.system.auxiliary.auxiliary import Auxiliary
from simses.system.power_electronics.acdc_converter.abstract_acdc_converter import AcDcConverter


class PowerElectronics:
    """ class to update the power elections values"""

    def __init__(self, acdc_converter: AcDcConverter):
        self.__log: Logger = Logger(type(self).__name__)
        self.__acdc_converter: AcDcConverter = acdc_converter

    def update(self, time: float, state: SystemState):
        """
        Function to update states regarding the power electronics unit

        Parameters
        ----------
        state :
            current SystemState

        Returns
        -------

        """
        ac_power = self.__check_power_limits(state)
        dc_power = self.get_dc_power_from(ac_power, state.dc_circuit_voltage)
        state.pe_losses = abs(ac_power - dc_power)
        state.dc_power_intermediate_circuit = dc_power

    def get_auxiliaries(self) -> [Auxiliary]:
        return list()

    @property
    def volume(self) -> float:
        """
        Volume of power electronics in m3

        Returns
        -------

        """
        volume: float = 0.0
        volume += self.__acdc_converter.volume
        return volume

    def get_dc_power_from(self, ac_power: float, voltage: float) -> float:
        """
        function to get the dc power from the power electronics unit

        Parameters
        ----------
        ac_power :
            requested ac power from the EMS
        voltage :
            DC voltage

        Returns
        -------
        float
            dc power
        """

        if self.__is_charge(ac_power):
            dc_power = self.__charge(ac_power, voltage)
        else:
            dc_power = self.__discharge(ac_power, voltage)
        return dc_power

    def update_ac_power_from(self, state: SystemState) -> None:
        """
        function to update ac power if fulfillment is not 100 %

        Parameters
        ----------
        state :
            current SystemState

        Returns
        -------

        """
        dc_power_delivered = state.dc_power_intermediate_circuit - state.dc_power_additional
        if self.__is_charge(dc_power_delivered):
            ac_power_converter = self.__acdc_converter.to_dc_reverse(dc_power_delivered, state.dc_circuit_voltage)
        else:
            ac_power_converter = self.__acdc_converter.to_ac_reverse(dc_power_delivered, state.dc_circuit_voltage)
        if abs(ac_power_converter) > self.max_power * 1.1:  # acceptable overload of 10%
            self.__log.error('Power electronics is not able to deliver full power requested of ' +
                             str(int(ac_power_converter)) + ' W (max. power: ' + str(int(self.max_power)) + ' W)')
        state.ac_power_delivered = ac_power_converter + state.aux_losses
        state.pe_losses = abs(ac_power_converter - dc_power_delivered)
        state.dc_power_intermediate_circuit = dc_power_delivered

    @staticmethod
    def __is_charge(power: float) -> bool:
        """
        Function to check, if the system is charged or discharged

        Parameters
        ----------
        power :
            ac power

        Returns
        -------
        bool
            true if the system is charged

        """
        return True if power > 0 else False

    def __charge(self, ac_power: float, voltage: float) -> float:
        """
        function to get the dc power from the power electronics unit if the system is charged

        Parameters
        ----------
        ac_power :
            requested ac power from the EMS
        voltage :
            DC voltage

        Returns
        -------
        float
            dc power
        """
        power_dc = self.__acdc_converter.to_dc(ac_power, voltage)
        return power_dc

    def __discharge(self, ac_power: float, voltage: float) -> float:
        """
        function to get the dc power from the power electronics unit if the system is discharged

        Parameters
        ----------
        ac_power :
            requested ac power from the EMS
        voltage :
            DC voltage

        Returns
        -------
        float
            dc power
        """

        power_dc = self.__acdc_converter.to_ac(ac_power, voltage)
        return power_dc

    def __check_power_limits(self, state: SystemState) -> float:
        """
        Checks if the requested ac power can be handled by the PE unit

        Parameters
        ----------
        state :
            current SystemState

        Returns
        -------
        float
            ac_power, which can be handled by the PE unit

        """
        # TODO Is this the right place for integrating aux losses?
        # Please see Issue 113 (https://gitlab.lrz.de/ees-ses/simses/-/issues/113)
        ac_power = state.ac_power
        if abs(ac_power) > self.__acdc_converter.standby_power_threshold:
            ac_power -= state.aux_losses
        if abs(ac_power) > self.__acdc_converter.max_power:
            self.__log.warn('Requested power (' + str(ac_power) + ') is beyond the maximum power of the '
                            'power electronics. Power is set to: '+ str(self.__acdc_converter.max_power))
            return np.sign(ac_power) * self.__acdc_converter.max_power
        else:
            return ac_power

    @property
    def max_power(self) -> float:
        return self.__acdc_converter.max_power

    @property
    def acdc_converter_type(self) -> str:
        """
        Returns the name of the selected acdc converter

        Parameters
        ----------

        Returns
        -------
        str
            name of the selected acdc converter

        """
        return type(self.__acdc_converter).__name__

    def close(self) -> None:
        """
        Closing all open resources in the power electronics unit

        Parameters
        ----------

        Returns
        -------

        """
        self.__acdc_converter.close()
        self.__log.close()



import numpy as np
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D
import math as m
import pandas as pd
import csv

__TEMPERATURE_CELL_C: float = 80  # °C
__TEMPERATURE_CELL_K: float = 273.15 + __TEMPERATURE_CELL_C  # K
__TEMPERATURE_REF_K: float = 298.15  # K

__IDEAL_GAS_CONST: float = 8.314462  # J/(mol K)
__FARADAY_CONST: float = 96485.3321  # As/mol

__CONTACT_RESISTANCE = 1.25 * 10 ** -3  # Ohm * cm2

__AREA_CELL = 600  # cm2
__MAX_CURRENT_DENSITY = 1  # A/cm2
__REV_VOLTAGE_STC = 1.229  # V  reversible Zellspannung unter STC
__ALPHA_ANODE = 0.25
__ALPHA_CATHODE = 0.41
__REF_EX_CURRENT_DENS_ANODE = 171 * 10 ** -3  # A/cm2
__REF_EX_CURRENT_DENS_CATHODE = 2.8 * 10 ** -6  # A/cm2
__ACT_ENERGY_ANODE = 24359  # J/mol
__ACT_ENERGY_CATHODE = 40000  # J/mol
__LAMBDA = 22.5  # degree of humidification
__THIKNESS_NAFION = 50.8  # um
__CONDUCTIVITY_NAFION = (0.005139 * __LAMBDA - 0.00326) * m.exp(
    1268 * (1 / 303 - 1 / __TEMPERATURE_CELL_K))  # S/cm 0.14
__mem_resistance = __THIKNESS_NAFION * 10 ** -4 / __CONDUCTIVITY_NAFION  # Ohm * cm2

__ATM = 1.013  # bar

__PRESSURE_ANODE = 1.7  # bar   total pressure Anode
__PRESSURE_CATHODE = 1.7  # bar   total pressure Cathode

__pressure_sat_h2o = 10 ** (
        -2.1794 + 0.02953 * __TEMPERATURE_CELL_C - 9.1837e-5 * __TEMPERATURE_CELL_C ** 2 + 1.4454e-7 *
        __TEMPERATURE_CELL_C ** 3)

__pressure_h2 = (__PRESSURE_ANODE - __pressure_sat_h2o) / __ATM  # bar to atm
__pressure_o2 = (__PRESSURE_CATHODE - __pressure_sat_h2o) * 0.21 / __ATM  # bar to atm

__open_circuit_voltage = __REV_VOLTAGE_STC + (__IDEAL_GAS_CONST * __TEMPERATURE_CELL_K / (2 * __FARADAY_CONST)) \
                         * m.log(__pressure_h2 * __pressure_o2 ** 0.5) - 0.85 * 10 ** -3 * (__TEMPERATURE_CELL_K
                                                                                            - __TEMPERATURE_REF_K)

__current_density = np.linspace(0, 1000, 500) * 10 ** -3  # mA

__ex_current_dens_anode = __REF_EX_CURRENT_DENS_ANODE * m.exp((-__ACT_ENERGY_ANODE / __IDEAL_GAS_CONST) * \
                                                              (1 / __TEMPERATURE_CELL_K - 1 / 333))  # 1 / 333 sollte eventuell 1/298.1 sein??, da bezogen auf STC

__ex_current_dens_cathode = __REF_EX_CURRENT_DENS_CATHODE * m.exp((-__ACT_ENERGY_CATHODE / __IDEAL_GAS_CONST) * \
                                                              (1 / __TEMPERATURE_CELL_K - 1 / 298.15))

_cell_voltage = np.ones(len(__current_density))

_spec_cell_power = np.ones(len(__current_density))

for i in range(0, len(__current_density)):
    _cell_voltage[i] = __open_circuit_voltage - __IDEAL_GAS_CONST * __TEMPERATURE_CELL_K / \
                       (2 * __ALPHA_ANODE * __FARADAY_CONST) * m.asinh(__current_density[i] / __ex_current_dens_anode) - \
                       __IDEAL_GAS_CONST * __TEMPERATURE_CELL_K / \
                       (2 * __ALPHA_CATHODE * __FARADAY_CONST) * m.asinh(__current_density[i] /
                                                                     __ex_current_dens_cathode) - __CONTACT_RESISTANCE \
                      * __current_density[i] - __mem_resistance * __current_density[i]

for i in range(0, len(__current_density)):
    _spec_cell_power[i] = _cell_voltage[i] * __current_density[i]

x = __current_density
y1 = _cell_voltage
y2 = _spec_cell_power

ax1 = plt.subplot(211)
plt.plot(x, y1)

ax2 = plt.subplot(212)
plt.plot(x, y2)
plt.show()

dict = {'currentdensity': __current_density, 'cellvoltage at 1 bar': _cell_voltage}
dict2 = {'currentdenstiy': __current_density, 'cellpower at 1 bar': _spec_cell_power}

# df = pd.DataFrame(dict)
# df2 = pd.DataFrame(dict2)

# df.to_csv('polarizationcurve1bar.csv', index=False)
# df2.to_csv('powercurve1bar.csv', index=False)



from simses.system.housing.forty_ft_container import FortyFtContainer
from simses.system.housing.twenty_ft_container import TwentyFtContainer
from simses.commons.data.data_handler import DataHandler
from simses.commons.error import EndOfLifeError
from simses.commons.state.abstract_state import State
from simses.commons.state.parameters import SystemParameters
from simses.commons.state.system import SystemState
from simses.logic.power_distribution.power_distributor import PowerDistributor
from simses.system.auxiliary.auxiliary import Auxiliary
from simses.system.dc_coupling.dc_coupler import DcCoupling
from simses.system.housing.abstract_housing import Housing
from simses.system.power_electronics.electronics import PowerElectronics
from simses.system.storage_system_dc import StorageSystemDC
from simses.system.thermal.model.system_thermal_model import SystemThermalModel


class StorageSystemAC:
    """
    AC storage system class manages all connections within the AC storage system. Connections are power electronics,
    further DC couplings (load/generation), housing, thermal model, auxiliaries, DC storage systems and its power
    distribution logic. DC couplings are handled as an interference and tried to be primarily fed by storage system
    power.
    """

    def __init__(self, state: SystemState, data_export: DataHandler, system_thermal_model: SystemThermalModel,
                 power_electronics: PowerElectronics, storage_systems: [StorageSystemDC], dc_couplings: [DcCoupling]
                 , housing: Housing, power_distributor: PowerDistributor):
        self.__data_export = data_export
        self.__system_state: SystemState = state
        self.__system_id: int = int(self.__system_state.get(SystemState.SYSTEM_AC_ID))
        self.__power_electronics: PowerElectronics = power_electronics
        self.__acdc_converter: str = self.__power_electronics.acdc_converter_type
        self.__system_thermal_model: SystemThermalModel = system_thermal_model
        self.__storage_systems: [StorageSystemDC] = storage_systems
        self.__dc_couplings: [DcCoupling] = dc_couplings
        self.__auxiliaries: [Auxiliary] = list()
        self.__auxiliaries.extend(self.__system_thermal_model.get_auxiliaries())
        self.__auxiliaries.extend(self.__power_electronics.get_auxiliaries())
        self.__housing = housing
        for storage_system in self.__storage_systems:
            self.__auxiliaries.extend(storage_system.get_auxiliaries())
            self.__housing.add_component_volume(storage_system.volume)
        for dc_coupling in self.__dc_couplings:
            self.__auxiliaries.extend(dc_coupling.get_auxiliaries())
        self.__system_state = self.__get_current_state()
        self.__data_export.transfer_data(self.__system_state.to_export())  # initial timestep
        self.__power_distributor: PowerDistributor = power_distributor
        self.__housing.add_component_volume(self.__power_electronics.volume)
        self.__system_thermal_model.update_air_parameters()

    def update(self, power: float, time: float) -> None:
        """
        Function to update the states of an AC storage system

        Parameters
        ----------
        power : ac target power (from energy management system)
        time : current simulation time

        Returns
        -------

        """
        state = self.__system_state
        state.ac_power = power
        # TODO Aux losses are calculated from last timestep. How to implement aux losses for current timestep?
        state.aux_losses = 0
        state.dc_power_additional = 0
        for aux in self.__auxiliaries:  # type: Auxiliary
            aux.update(time, state)
        for dc in self.__dc_couplings:  # type: DcCoupling
            dc.update(time, state)
        self.__power_electronics.update(time, state)

        states: [State] = list()
        for system in self.__storage_systems:  # type: StorageSystemDC
            states.append(system.state)

        # self.__system_thermal_model.calculate_temperature(time, state, states)

        self.__power_distributor.set(time, states, power)

        total_dc_power: float = state.dc_power_intermediate_circuit + state.dc_power_additional
        all_systems_ok: bool = True
        err_msg: str = ''
        for system in self.__storage_systems:  # type: StorageSystemDC
            # TODO how to decide whether to store additional dc load or generation? (MM)
            local_power: float = self.__power_distributor.get_power_for(total_dc_power, system.state)
            try:
                system.update(time, local_power)
            except EndOfLifeError as err:
                all_systems_ok = False
                err_msg = str(err)
        for system in self.__storage_systems:  # type: StorageSystemDC
            system.wait()
        state.time = time
        self.__system_state = self.__get_current_state()
        # Make sure to run the temperature model after running the DC System to represent ACDC losses for corner cases
        # of SOC=0 and SOC=1 correctly
        self.__system_thermal_model.calculate_temperature(time, self.__system_state, states)
        self.__data_export.transfer_data(self.__system_state.to_export())
        if not all_systems_ok:
            raise EndOfLifeError(err_msg)

    def __get_current_state(self) -> SystemState:
        """
        Calculation of current system state

        Returns
        -------
        SystemState:
            Current state of system
        """
        states = list()
        for system in self.__storage_systems:  # type: StorageSystemDC
            states.append(system.state)
        system_state: SystemState = SystemState.sum_parallel(states)
        system_state.set(SystemState.SYSTEM_AC_ID, self.__system_id)
        system_state.set(SystemState.SYSTEM_DC_ID, 0)
        system_state.time = self.__system_state.time
        system_state.temperature = self.__system_thermal_model.get_temperature()
        if isinstance(self.__housing, (TwentyFtContainer, FortyFtContainer)):
            system_state.ol_temperature = self.__system_thermal_model.get_ol_temperature()
            system_state.il_temperature = self.__system_thermal_model.get_il_temperature()
        system_state.ambient_temperature = self.__system_thermal_model.get_ambient_temperature()
        system_state.solar_thermal_load = self.__system_thermal_model.get_solar_irradiation_thermal_load()
        system_state.hvac_thermal_power = self.__system_thermal_model.get_hvac_thermal_power()
        system_state.ac_power = self.__system_state.ac_power
        system_state.pe_losses = self.__system_state.pe_losses
        system_state.aux_losses = self.__system_state.aux_losses
        system_state.dc_power_additional = self.__system_state.dc_power_additional
        system_state.max_charge_power = min(system_state.max_charge_power, self.__power_electronics.max_power)
        system_state.max_discharge_power = min(system_state.max_discharge_power, self.__power_electronics.max_power)
        self.__power_electronics.update_ac_power_from(system_state)
        # Include power electronic fulfillment after reverse power calculation
        if abs(system_state.ac_power) < max(system_state.max_charge_power, system_state.max_discharge_power) * 0.01:
            system_state.ac_fulfillment = 1
        else:
            system_state.ac_fulfillment = system_state.ac_power_delivered / system_state.ac_power
        return system_state

    def get_system_parameters(self) -> dict:
        parameters: dict = dict()
        parameters[SystemParameters.SYSTEM] = type(self).__name__
        parameters[SystemParameters.ID] = str(self.__system_id)
        parameters[SystemParameters.POWER_DISTRIBUTION] = type(self.__power_distributor).__name__
        parameters[SystemParameters.CONTAINER_NUMBER] = self.__housing.get_number_of_containers()
        parameters[SystemParameters.CONTAINER_TYPE] = type(self.__housing).__name__
        parameters[SystemParameters.ACDC_CONVERTER] = self.__acdc_converter
        auxiliaries: list = list()
        for aux in self.__auxiliaries:  # type: Auxiliary
            auxiliaries.append(type(aux).__name__)
        subsystems: list = list()
        for system in self.__storage_systems:  # type: StorageSystemDC
            subsystems.append(system.get_system_parameters())
        parameters[SystemParameters.AUXILIARIES] = auxiliaries
        parameters[SystemParameters.SUBSYSTEM] = subsystems
        return parameters

    @property
    def state(self) -> SystemState:
        """
        Returns current system state

        Returns
        -------
        SystemState:
            State of the ac system

        """
        return self.__system_state

    @property
    def max_power(self) -> float:
        return self.__power_electronics.max_power

    @property
    def system_id(self) -> int:
        return self.__system_id

    def reset_profiles(self, ts_adapted: float) -> None:
        """
        Enables looping of the simulation beyond the original length of the time series for the AmbientThermalModel and
        SolarIrradiationModel
        """
        self.__system_thermal_model.reset_profiles(ts_adapted)

    def close(self) -> None:
        """
        Closing all open resources in AC storage system

        Parameters
        ----------

        Returns
        -------

        """
        self.__power_electronics.close()
        self.__system_thermal_model.close()
        self.__power_distributor.close()
        for storage_system in self.__storage_systems:  # type: StorageSystemDC
            storage_system.close()
        for aux in self.__auxiliaries:  # type: Auxiliary
            aux.close()
        for dc_coupling in self.__dc_couplings:  # type: DcCoupling
            dc_coupling.close()



import pandas
from simses.analysis.data.abstract_data import Data
from simses.commons.config.simulation.general import GeneralSimulationConfig
from simses.commons.state.technology.lithium_ion import LithiumIonState


class LithiumIonData(Data):
    """
    Provides time series data from LithiumIonState
    """

    def __init__(self, config: GeneralSimulationConfig, data: pandas.DataFrame):
        super().__init__(config, data)

    @property
    def id(self) -> str:
        return str(int(self._get_first_value(LithiumIonState.SYSTEM_AC_ID))) + '.' + \
               str(int(self._get_first_value(LithiumIonState.SYSTEM_DC_ID)))

    @property
    def time(self):
        return self._get_data(LithiumIonState.TIME)

    @property
    def power(self):
        return self.voltage * self.current

    @property
    def current(self):
        return self._get_data(LithiumIonState.CURRENT)

    @property
    def voltage(self):
        return self._get_data(LithiumIonState.VOLTAGE)

    @property
    def dc_power(self):
        return self.power

    @property
    def energy_difference(self):
        # initial_soc = self._get_first_value(LithiumIonState.SOC)
        # last_soc = self._get_last_value(LithiumIonState.SOC)
        # initial_capacity = float(self.capacity[0])
        # last_capacity = float(self.capacity[-1])
        # return last_soc * last_capacity - initial_soc * initial_capacity
        initial_soe = float(self.soe[0])
        last_soe = float(self.soe[-1])
        return last_soe - initial_soe

    @property
    def soe(self):
        return self._get_data(LithiumIonState.SOE) / 1000.0

    @property
    def soc(self):
        return self._get_data(LithiumIonState.SOC)

    @property
    def capacity(self):
        return self._get_data(LithiumIonState.CAPACITY) / 1000.0

    @property
    def nominal_voltage(self):
        return self._get_data(LithiumIonState.NOMINAL_VOLTAGE)[0]

    @property
    def state_of_health(self):
        return self._get_data(LithiumIonState.SOH)

    @property
    def resistance(self):
        return self._get_data(LithiumIonState.INTERNAL_RESISTANCE)

    @property
    def resistance_increase(self):
        return self._get_data(LithiumIonState.RESISTANCE_INCREASE)

    @property
    def storage_fulfillment(self):
        return self._get_data(LithiumIonState.FULFILLMENT)

    @property
    def temperature(self):
        return self._get_data(LithiumIonState.TEMPERATURE)

    @property
    def capacity_loss_calendar(self):
        return self._get_data(LithiumIonState.CAPACITY_LOSS_CALENDRIC)

    @property
    def capacity_loss_cyclic(self):
        return self._get_data(LithiumIonState.CAPACITY_LOSS_CYCLIC)

    @property
    def resistance_increase_calendar(self):
        return self._get_data(LithiumIonState.RESISTANCE_INCREASE_CALENDRIC)

    @property
    def resistance_increase_cyclic(self):
        return self._get_data(LithiumIonState.RESISTANCE_INCREASE_CYCLIC)

    @classmethod
    def get_system_data(cls, path: str, config: GeneralSimulationConfig) -> list:
        system_data: [pandas.DataFrame] = cls._get_system_data_for(path, LithiumIonState, LithiumIonState.TIME,
                                                                   LithiumIonState.SYSTEM_AC_ID, LithiumIonState.SYSTEM_DC_ID)
        res: [LithiumIonData] = list()
        for data in system_data:
            res.append(LithiumIonData(config, data))
        return res





from configparser import ConfigParser

from simses.commons.config.abstract_config import Config
from simses.commons.utils.utilities import get_path_for


class DataConfig(Config):

    """
    DataConfig objects provide information for each specific system path to data
    """

    CONFIG_NAME: str = 'data'
    CONFIG_PATH: str = get_path_for(CONFIG_NAME + Config.DEFAULTS)

    def __init__(self, path: str, config: ConfigParser):
        if path is None:
            path = self.CONFIG_PATH
        super().__init__(path, self.CONFIG_NAME, config)



from configparser import ConfigParser

from simses.commons.data.data_handler import DataHandler
from simses.commons.state.energy_management import EnergyManagementState
from simses.commons.state.system import SystemState
from simses.logic.energy_management.energy_management_factory import EnergyManagementFactory
from simses.logic.energy_management.strategy.operation_strategy import OperationStrategy


class EnergyManagement:
    """
    Energy Management System to control the operation of the ESS.
    """

    def __init__(self, data_export: DataHandler, config: ConfigParser):
        self.__config: ConfigParser = config
        self.__factory: EnergyManagementFactory = EnergyManagementFactory(config)
        self.__operation_strategy: OperationStrategy = self.__factory.create_operation_strategy()
        self.__state: EnergyManagementState = self.__factory.create_energy_management_state()
        self.__data_export: DataHandler = data_export
        self.__data_export.transfer_data(self.__state.to_export())  # initial timestep

    def next(self, time: float, system_state: SystemState, power_offset: float = 0) -> float:
        """
        Return the next power value of the energy management strategy in W based on the operation strategy.

        Parameters
        ----------
        time : float
            Current timestamp.
        time_loop_delta:float
            Time delta due to current simulation loop.
        system_state : State
            Current state of the system.
        power : float
            Power value of stacked operation strategy.

        Returns
        -------
        float:
            Power value the StorageSystem should meet in W.

        """
        power = self.__operation_strategy.next(time, system_state, power_offset)
        self.__operation_strategy.update(self.__state)
        self.__state.ac_power_requested = power
        # Substract time loop delta for reading from profile csv
        # self.__state.time = time - time_loop_delta
        return power

    def export(self, time: float) -> None:
        self.__state.time = time
        self.__data_export.transfer_data(self.__state.to_export())

    def create_instance(self):
        return EnergyManagement(self.__data_export, self.__config)

    def close(self) -> None:
        """Closing open resources"""
        self.__operation_strategy.close()
        self.__factory.close()



import numpy as np
from scipy import optimize

from simses.commons.log import Logger
from simses.commons.utils.utilities import check
from simses.system.power_electronics.acdc_converter.abstract_acdc_converter import AcDcConverter


class SungrowAcDcConverter(AcDcConverter):
    """  Fitted Efficiency Curve using field data: Master Thesis by Felix Müller"""

    """ Source:
        Field data from fcr storage system. Datasheet of converter can be found here:
        ...\simses\simulation\storage_system\power_electronics\acdc_converter\data\datasungrow_SC1000TL_spec.pdf
    """

    __fit_dch = 1 # 1: Notton, 2: Rampinelli, 3: Rational12
    __fit_ch = 1 # 1: Notton, 2: Rampinelli, 3: Rational12
    __min_eff = 0.2092 # Minimum for efficiency (in case of Discharge and Notton or Rampinelli is used)

    # Notton
    # Fit for Discharging
    __P0_to_ac_Notton = 0.005511580
    __K_to_ac_Notton = 0.018772838
    # Fit for Charging
    __P0_to_dc_Notton = 0.007701864
    __K_to_dc_Notton = 0.017290859

    # Rampinelli
    # Fit for Discharging
    __K0_to_ac_Rampinelli = 0.003407887
    __K1_to_ac_Rampinelli = 0.013809826
    __K2_to_ac_Rampinelli = 0.003155305
    # Fit for Charging
    __K0_to_dc_Rampinelli = 0.007421847
    __K1_to_dc_Rampinelli = 0.003452202
    __K2_to_dc_Rampinelli = 0.011994448

    # Rational12
    # Fit for Discharging
    __a1_to_ac_Rat = 57.341420538
    __a0_to_ac_Rat = 0.092381040
    __b1_to_ac_Rat = 57.318868901
    __b0_to_ac_Rat = 0.441493908
    # Fit for Charging
    __a1_to_dc_Rat = 47.773200770
    __a0_to_dc_Rat = 0.210333852
    __b1_to_dc_Rat = 47.572383928
    __b0_to_dc_Rat = 0.630988885

    __VOLUMETRIC_POWER_DENSITY = 3.227 * 1e5  # W / m3
    # Calculated Value: Nominal Power / (W * H * D)
    __GRAVIMETRIC_POWER_DENSITY = 909  # W/kg
    # Calculated Value: Nominal Power / Weight
    __SPECIFIC_SURFACE_AREA = 1.349 * 1e-5  # in m2 / W
    # Calculated Value: (W*H + W*D + H*D)*2 / Nominal Power

    def __init__(self, max_power: float):
        super().__init__(max_power)
        self.__log: Logger = Logger(type(self).__name__)

    def to_ac(self, power: float, voltage: float) -> float:
        check(power)
        res: float = 0.0
        if power < 0.0:
            res = power / self.__get_efficiency_to_ac(power)
        return res

    def to_dc(self, power: float, voltage: float) -> float:
        check(power)
        res: float = 0.0
        if power > 0.0:
            res = power * self.__get_efficiency_to_dc(power)
        return res

    def to_dc_reverse(self, dc_power: float, voltage: float) -> float:
        check(dc_power)
        res: float = 0.0
        if dc_power > 0.0:
            if self.__fit_ch == 1:
                p = - dc_power / (1 - dc_power * self.__K_to_dc_Notton / self.max_power)
                q = - self.__P0_to_dc_Notton * dc_power / \
                    (1 / self.max_power - abs(dc_power) * self.__K_to_dc_Notton / self.max_power ** 2)
            elif self.__fit_ch == 2:
                p = - dc_power * (self.__K1_to_dc_Rampinelli + 1) / (
                        1 - dc_power * self.__K2_to_dc_Rampinelli / self.max_power)
                q = - self.__K0_to_dc_Rampinelli * dc_power / (
                        1 / self.max_power - abs(dc_power) * self.__K_to_dc_Notton / self.max_power ** 2)
            elif self.__fit_ch == 3:
                p = (self.__a0_to_dc_Rat * self.max_power - self.__b1_to_dc_Rat * dc_power) / (
                            self.__a1_to_dc_Rat - (dc_power / self.max_power))
                q = -(dc_power * self.__b0_to_dc_Rat * self.max_power) / (self.__a1_to_dc_Rat - (dc_power / self.max_power))
            else:
                raise Exception('Fit for Charge is not possible: __fit_ch=' + str(self.__fit_ch))
            self.__log.debug('P_DC: ' + str(dc_power))
            res = max(0.0, -p / 2 + np.sqrt((p / 2) ** 2 - q))
        return res

    def to_ac_reverse(self, dc_power: float, voltage: float) -> float:
        check(dc_power)
        res: float = 0.0
        if dc_power < 0.0:
            if self.__fit_dch == 1:
                p = self.max_power / self.__K_to_ac_Notton
                q = (self.__P0_to_ac_Notton * self.max_power ** 2 -
                     abs(dc_power) * self.max_power) / self.__K_to_ac_Notton
                power_ac = min(0.0, -(-p / 2 + np.sqrt((p / 2) ** 2 - q)))
                res = min(power_ac, self.__min_eff * dc_power)
            elif self.__fit_dch == 2:
                p = (1 + self.__K1_to_ac_Rampinelli) * self.max_power / self.__K0_to_ac_Rampinelli
                q = (self.__K0_to_ac_Rampinelli * self.max_power ** 2 - abs(dc_power) * self.max_power) / \
                    self.__K2_to_ac_Rampinelli
                power_ac = min(0.0, -(-p / 2 + np.sqrt((p / 2) ** 2 - q)))
                res = min(power_ac, self.__min_eff * dc_power)
            elif self.__fit_dch == 3:
                def fun(x):
                    return x ** 3 + self.max_power * self.__b1_to_ac_Rat * x ** 2 + \
                           (self.__b0_to_ac_Rat * self.max_power ** 2 - self.max_power * self.__a1_to_ac_Rat * abs(
                               dc_power)) * x - \
                           self.max_power ** 2 * self.__a0_to_ac_Rat * abs(dc_power)
                res = optimize.root(fun, abs(dc_power)).x[0] * (-1)
            else:
                raise Exception('Fit for Discharge is not possible: __fit_dch=' + str(self.__fit_dch))
        return res

    def __get_efficiency_to_ac(self, power: float) -> float:
        power_factor = abs(power) / self.max_power
        if power_factor < 0.0 or power_factor > 1.0:
            raise Exception('Power factor is not possible: ' + str(power_factor))
        if self.__fit_dch == 1:
            return max(self.__min_eff, power_factor /
                       (power_factor + self.__P0_to_ac_Notton + self.__K_to_ac_Notton * power_factor ** 2))
        elif self.__fit_dch == 2:
            return max(self.__min_eff, power_factor / (power_factor + (self.__K0_to_ac_Rampinelli +
                       self.__K1_to_ac_Rampinelli * power_factor + self.__K2_to_ac_Rampinelli * power_factor ** 2)))
        elif self.__fit_dch == 3:
            return (self.__a1_to_ac_Rat * power_factor + self.__a0_to_ac_Rat) / \
                   (power_factor ** 2 + self.__b1_to_ac_Rat * power_factor + self.__b0_to_ac_Rat)
        else:
            raise Exception('Fit for Discharge is not possible: __fit_dch=' + str(self.__fit_dch))

    def __get_efficiency_to_dc(self, power: float) -> float:
        power_factor = abs(power) / self.max_power
        if power_factor < 0.0 or power_factor > 1.0:
            raise Exception('Power factor is not possible: ' + str(power_factor))
        if self.__fit_ch == 1:
            return power_factor / (power_factor + self.__P0_to_dc_Notton + self.__K_to_dc_Notton * power_factor ** 2)
        elif self.__fit_ch == 2:
            return power_factor / (power_factor + (self.__K0_to_dc_Rampinelli +
                       self.__K1_to_dc_Rampinelli * power_factor + self.__K2_to_dc_Rampinelli * power_factor ** 2))
        elif self.__fit_ch == 3:
            return (self.__a1_to_dc_Rat * power_factor + self.__a0_to_dc_Rat) / \
                   (power_factor ** 2 + self.__b1_to_dc_Rat * power_factor + self.__b0_to_dc_Rat)
        else:
            raise Exception('Fit for Charge is not possible: __fit_ch=' + str(self.__fit_ch))

    @property
    def volume(self) -> float:
        return self.max_power / self.__VOLUMETRIC_POWER_DENSITY\

    @property
    def mass(self):
        return self.max_power / self.__GRAVIMETRIC_POWER_DENSITY

    @property
    def surface_area(self) -> float:
        return self.max_power * self.__SPECIFIC_SURFACE_AREA

    @classmethod
    def create_instance(cls, max_power: float, power_electronics_config=None):
        return SungrowAcDcConverter(max_power)

    def close(self) -> None:
        self.__log.close()



from abc import ABC, abstractmethod

from simses.commons.state.energy_management import EnergyManagementState
from simses.commons.state.system import SystemState


class OperationStrategy(ABC):
    """
    Algorithm to determine output power of the ESS for the next timestep.
    """

    def __init__(self, priority: float):
        super().__init__()
        self.__priority = priority

    @abstractmethod
    def next(self, time: float, system_state: SystemState, power: float = 0) -> float:
        """
        Provides the next value for the output power of the ESS.

        Parameters
        ----------
        time : float
        system_state : SystemState
        power : float
            Power value for stacked operation strategy.

        Returns
        -------
        float
            Output power for the ESS in the next timestep.
        """
        pass

    @abstractmethod
    def update(self, energy_management_state: EnergyManagementState) -> None:
        """
        In-place update of energy management state variables

        Parameters
        ----------
        energy_management_state : state for energy management

        Returns
        -------

        """
        pass

    @abstractmethod
    def clear(self) -> None:
        """
        Clearing internal values of operations strateggy

        Returns
        -------

        """
        pass

    @abstractmethod
    def close(self) -> None:
        """Closing open resources"""
        pass

    @property
    def priority(self):
        """Defines the priority of the operation strategy in a multiuse case
        VERY_HIGH = 0
        HIGH = 1
        MEDIUM = 2
        LOW = 3
        VERY_LOW = 4"""
        return self.__priority

    @staticmethod
    def sort(strategies: []) -> None:
        """Sorts the order in which the operation strategies are applied in a multiuse case"""
        strategies.sort(key=lambda strategy: strategy.priority, reverse=True)



from abc import ABC, abstractmethod

from simses.commons.state.technology.storage import StorageTechnologyState
from simses.system.auxiliary.auxiliary import Auxiliary


class StorageTechnology(ABC):
    """Abstract class for all technologies"""

    def __init__(self):
        pass

    @abstractmethod
    def distribute_and_run(self, time: float, current: float, voltage: float):
        """
        starts the update process of a data

        Parameters
        ----------
        time : current simulation time
        current : requested current for a data
        voltage : dc voltage for a data

        Returns
        -------

        """
        pass

    @abstractmethod
    def wait(self):
        pass

    @abstractmethod
    def get_auxiliaries(self) -> [Auxiliary]:
        pass

    @property
    @abstractmethod
    def state(self) -> StorageTechnologyState:
        """
        function to get the data state

        Parameters
        -------

        Returns
        -------
            StorageTechnologyState : specific data state

        """
        pass

    @property
    @abstractmethod
    def volume(self) -> float:
        """
        Volume of storage technology in m3

        Returns
        -------

        """
        pass

    @property
    @abstractmethod
    def mass(self) -> float:
        """
        Mass of storage technology in kg

        Returns
        -------

        """
        pass

    @property
    @abstractmethod
    def surface_area(self) -> float:
        """
        Surface area of storage technology in m2

        Returns
        -------

        """
        pass

    @property
    @abstractmethod
    def specific_heat(self) -> float:
        """
               Specific heat of storage technology in J/(kgK)

               Returns
               -------

               """
        pass

    @property
    @abstractmethod
    def convection_coefficient(self) -> float:
        """
                determines the convective heat transfer coefficient of a battery cell

                Returns
                -------
                float:
                    convective heat transfer coefficient in W/(m^2*K)
                """
        pass

    @abstractmethod
    def get_system_parameters(self) -> dict:
        """
        All system parameters inherited by the storage technology

        Returns
        -------

        """
        pass

    @abstractmethod
    def close(self):
        """
        Closing all open resources in a data

        Parameters
        ----------

        Returns
        -------

        """
        pass


import pytest
from simses.commons.timeseries.interpolation.linear_interpolation import LinearInterpolation
from simses.commons.timeseries.timevalue import TimeValue


@pytest.fixture()
def uut():
    return LinearInterpolation()


@pytest.mark.parametrize('time, recent_time, recent_value, last_time, last_value, result',
                         [
                             (0, 0, 0, 10, 10, 0),
                             (10, 0, 0, 10, 10, 10),
                             (5, 0, 0, 10, 10, 5)
                         ]
                         )
def test_next(time, recent_time, recent_value, last_time, last_value, result, uut):
    recent = TimeValue(tstmp=recent_time, value=recent_value)
    last = TimeValue(tstmp=last_time, value=last_value)
    assert abs(uut.interpolate(time, recent, last) - result) < 1e-10



from configparser import ConfigParser

from simses.commons.config.simulation.simulation_config import SimulationConfig


class ElectrolyzerConfig(SimulationConfig):
    """
    Electrolyzer specific configs
    """

    def __init__(self, config: ConfigParser, path: str = None):
        super().__init__(path, config)
        self.__section: str = 'ELECTROLYZER'

    @property
    def eol(self) -> float:
        """Returns end of life criterion from data_config file_name"""
        return float(self.get_property(self.__section, 'EOL'))

    @property
    def desire_pressure_cathode(self) -> float:
        """Retruns desired pressure of cathode of electrolyzer"""
        return float(self.get_property(self.__section, 'PRESSURE_CATHODE'))

    @property
    def desire_pressure_anode(self) -> float:
        """Retruns desired pressure of cathode of electrolyzer"""
        return float(self.get_property(self.__section, 'PRESSURE_ANODE'))

    @property
    def desire_temperature(self) -> float:
        """Retruns desired pressure of cathode of electrolyzer"""
        return float(self.get_property(self.__section, 'TEMPERATURE'))



from matplotlib import pyplot
from numpy import ndarray, linspace, inf, mean
from pandas import DataFrame, read_csv
from scipy.optimize import curve_fit

bounds_for_custom_fitting_parameters = ([0, -inf, -inf], [inf, inf, inf])


def custom_fitting_function(x, b, c, d):
    return x / (x + b + c*x + d*x**2)


def polynomial_fit(x, *args):
    res = 0
    for order in range(len(args)):
        res += args[order] * x**order
    return res


# def multi_dimensional_fitting_function(X, a, b, c):
#     x,y = X
#     return np.log(a) + b*np.log(x) + c*np.log(y)

sorted = linspace(0.0, 1.0, 100)


class AixAcDcFitter:

    __FILE_NAME: str = './efficiencies.csv'
    __POWER_COL: str = 'Power'
    __EFFICIENCY_COL: str= 'Efficiency'
    __MODULE_COL: str = 'Module'
    __BLOCK_COL: str = 'Block'
    __DIRECTION_COL: str = 'Direction'
    __RESULT_FILE: str = './fitting.csv'

    __CHARGE: str = 'Quelle'
    __DISCHARGE: str = 'Senke'

    __MAX_POWER: float = 166.0  # kW

    __DO_PLOT: bool = True

    def __init__(self, block: int, module: int):
        data: DataFrame = read_csv(self.__FILE_NAME, sep=';', decimal=',')
        filtered_data: DataFrame = self.__filter(data=data, block=block, module=module)
        filtered_power: DataFrame = self.__filter_power_range(data=filtered_data, block=block, module=module)
        self.__power_data = filtered_power[self.__POWER_COL].to_numpy()
        self.__efficiency_data: ndarray = filtered_power[self.__EFFICIENCY_COL].to_numpy() / 100.0
        if self.__DO_PLOT:
            fig, self.__ax = pyplot.subplots()
            self.__ax.plot(self.__power_data, self.__efficiency_data, label='measured', marker='o', linestyle='none')

    def __filter(self, data: DataFrame, block: int, module: int) -> ndarray:
        filtered_data: DataFrame = data.loc[(data[self.__BLOCK_COL] == block) &
                                            (data[self.__MODULE_COL] == module) &
                                            (data[self.__EFFICIENCY_COL] > 0) &
                                            (data[self.__POWER_COL] > 0), :].copy().reset_index()
        return filtered_data

    def __normalize_power_of(self, data: DataFrame) -> None:
        power_data: ndarray = data[self.__POWER_COL]
        data[self.__POWER_COL] = power_data / self.__MAX_POWER

    def __filter_power_range(self, data: DataFrame, block: int, module: int) -> DataFrame:
        self.__normalize_power_of(data)
        power_scale: float = block * module
        min_power: float = 0.02  # max(0.02, min(0.5, 0.25 * (power_scale - 1)))
        max_power: float = 1.0  # 0.25 * power_scale
        filtered_data: DataFrame = data.loc[(data[self.__POWER_COL] > min_power) &
                                            (data[self.__POWER_COL] < max_power), :].copy().reset_index()
        return filtered_data

    def fit_polynomial_function(self, order: int) -> ndarray:
        popt, pcov = curve_fit(polynomial_fit, xdata=self.__power_data, ydata=self.__efficiency_data, p0=[0] * (order + 1))
        return polynomial_fit(sorted, *popt)

    def fit_order_range(self, start: int, end: int, step: int) -> None:
        for order in range(start, end, step):
            if self.__DO_PLOT:
                fit_data = self.fit_polynomial_function(order)
                self.__ax.plot(sorted, fit_data, label='polyn. fit, order ' + str(order), linestyle='--')

    def fit_custom_function(self):
        optimal_fit, variance_fit = curve_fit(custom_fitting_function, xdata=self.__power_data, ydata=self.__efficiency_data,
                                              bounds=bounds_for_custom_fitting_parameters, maxfev=10000)
        mean_error, max_error = self.get_mean_standard_error_from(optimal_fit)
        print(mean_error, max_error, optimal_fit)
        if self.__DO_PLOT:
            fit_data = custom_fitting_function(sorted, *optimal_fit)
            self.__ax.plot(sorted, fit_data, label='custom fit', linestyle='--')
        return optimal_fit

    def get_mean_standard_error_from(self, optimal_fit) -> (float, float):
        fit = custom_fitting_function(self.__power_data, *optimal_fit)
        error = abs(self.__efficiency_data - fit)
        return mean(error), max(error)

    def fit(self) -> None:
        # self.fit_order_range(4, 5, 1)
        optimal_fit = self.fit_custom_function()
        self.show()
        return optimal_fit

    def show(self) -> None:
        if self.__DO_PLOT:
            self.__ax.legend(loc='lower right')
            # plt.savefig(self.__path + self.alphanumerize(self.__title) + '.pdf')
            pyplot.show()


if __name__ == '__main__':
    optimal_fits: list = list()
    for module in range(1, 3, 1):
        for block in range(1,3,1):
            print('Fitting for block ' + str(block) + ' and module ' + str(module))
            fitter: AixAcDcFitter = AixAcDcFitter(block=block, module=module)
            optimal_fits.append(fitter.fit())
    fig, ax = pyplot.subplots()
    count: int = 0
    for optimal_fit in optimal_fits:
        count += 1
        fit_data = custom_fitting_function(sorted, *optimal_fit)
        ax.plot(sorted, fit_data, label='custom fit ' + str(count), linestyle='--')
    ax.legend(loc='lower right')
    pyplot.show()



from abc import ABC, abstractmethod

from simses.commons.state.technology.electrolyzer import ElectrolyzerState


class CyclicDegradationModel(ABC):
    """
    Degradation Model for the cyclic aging of the Electrolyzer.
    """

    def __init__(self):
        super().__init__()

    def calculate_degradation(self, time: float, state: ElectrolyzerState):
        self.calculate_resistance_increase(time, state)
        self.calculate_exchange_current_dens_decrerase(state)

    @abstractmethod
    def calculate_resistance_increase(self, time: float, state: ElectrolyzerState) -> None:
        """
        update the cyclic internal resistance increase of a electrolyzer
        Parameters
        ----------
            time: Float
            state : HydrogenState

        Returns
        -------

        """
        pass

    @abstractmethod
    def calculate_exchange_current_dens_decrerase(self, state: ElectrolyzerState):
        """
        update the cyclic decrease of the exchange current denisty of the electrolyzer
        :param state:
        :return:
        """
        pass

    @abstractmethod
    def get_resistance_increase(self) -> float:
        """
        get the updated calendary resistance increase
        Returns
        -------
        float:
            resistance increase in [p.u.]
        """
        pass

    @abstractmethod
    def get_exchange_current_dens_decrease(self) -> float:
        """
        get the updated caledric exchange current density decrease
        :return:
        """

    @abstractmethod
    def reset(self, hydrogen_state: ElectrolyzerState) -> None:
        """
        resets all values within a calendar degradation model, if electrolyzer is replaced
        Parameters
        ----------
            hydrogen_state : HydrogenState

        Returns
        -------
        """
        pass

    @abstractmethod
    def close(self) -> None:
        """
        Closing all resources in calendar degradation model

        Returns
        -------

        """
        pass



from simses.commons.config.simulation.system import StorageSystemConfig
from simses.system.housing.abstract_housing import Housing
from simses.system.housing.layer import Layer


class FortyFtContainer(Housing):

    def __init__(self, housing_configuration: list, temperature: float = None):

        # Get optional user-defined values required for this class or use default values
        try:
            high_cube: bool = bool(housing_configuration[StorageSystemConfig.HOUSING_HIGH_CUBE])
        except (ValueError, IndexError, TypeError):
            high_cube = False
        try:
            azimuth: float = float(housing_configuration[StorageSystemConfig.HOUSING_AZIMUTH])
        except (ValueError, IndexError, TypeError):
            azimuth = 0
        try:
            outer_layer_absorptivity: float = float(housing_configuration[StorageSystemConfig.HOUSING_ABSORPTIVITY])
        except (ValueError, IndexError, TypeError):
            outer_layer_absorptivity = 0.15
        try:
            ground_albedo: float = float(housing_configuration[StorageSystemConfig.HOUSING_GROUND_ALBEDO])
        except (ValueError, IndexError, TypeError):
            ground_albedo = 0.2

        if temperature is None:
            temperature = 298.15  # default value = 25 °C

        # Container external dimensions (from mtcontainer)
        outer_length = 12.192  # m
        outer_breadth = 2.438  # m
        if high_cube:
            outer_height = 2.891  # m
        else:
            outer_height = 2.591  # m

        # Wall layer thicknesses
        thickness_inner_layer = 0.001  # m, L1 = layer 1 - Aluminium
        thickness_mid_layer = 0.050  # m, L2 = Layer 2 - Rock wool
        thickness_outer_layer = 0.0016  # m, L3 = Layer 3 - steel (1.6 mm)
        depth_corrugation = 0.03  # m (0.254)

        # Container internal dimensions
        inner_length = outer_length - 2 * (thickness_inner_layer + thickness_mid_layer + thickness_outer_layer + depth_corrugation)  # in m
        inner_breadth = outer_breadth - 2 * (thickness_inner_layer + thickness_mid_layer + thickness_outer_layer + depth_corrugation)  # m
        inner_height = outer_height - 2 * (thickness_inner_layer + thickness_mid_layer + thickness_outer_layer + depth_corrugation)  # m

        # Unavailable space: for gangways and internal access
        gangway_width = 0.8  # m
        unavailable_area = gangway_width * inner_length + gangway_width * inner_breadth - gangway_width * gangway_width  # m2
        unavailable_volume = unavailable_area * inner_height  # m3

        # Densities of layers
        density_inner_layer_material = 2700  # kg/m3 (for Aluminium)
        density_mid_layer_material = 100  # kg/m3 (for Rock wool)
        density_outer_layer_material = 8050  # kg/m3 (for Steel)

        # Container material thermal characteristics
        thermal_conductivity_inner_layer = 237  # W/mK  (for Aluminium)
        thermal_conductivity_mid_layer = 0.050  # W/mK (for Rock wool)
        thermal_conductivity_outer_layer = 14.4  # W/mK (for Steel - Stainless, Type 304)
        specific_heat_inner_layer = 910  # J/kgK (for Aluminium)
        specific_heat_mid_layer = 840  # J/kgK (for Rock wool)
        specific_heat_outer_layer = 500  # J/kgK (for Steel - Stainless, Type 304)
        convection_coefficient_air_inner_layer = 30  # W/m2K, Convection coefficient for convection from air to L1 material (exemplary value)
        convection_coefficient_air_outer_layer = 30  # W/m2K, Convection coefficient for convection from L3 material to air (exemplary value)

        # Container Orientation and Surroundings
        self.__azimuth = azimuth
        self.__albedo = ground_albedo

        # Initialize layer properties

        # Layer 1 attributes (inner layer - Aluminium)
        inner_layer_attributes = dict()
        inner_layer_attributes[Layer.LENGTH] = inner_length  # m
        inner_layer_attributes[Layer.BREADTH] = inner_breadth  # m
        inner_layer_attributes[Layer.HEIGHT] = inner_height  # m
        inner_layer_attributes[Layer.THICKNESS] = thickness_inner_layer  # m
        inner_layer_attributes[Layer.DENSITY] = density_inner_layer_material  # kg/m3 (for Aluminium)
        inner_layer_attributes[Layer.THERMAL_CONDUCTIVITY] = thermal_conductivity_inner_layer  # W/mK  (for Aluminium)
        inner_layer_attributes[Layer.SPECIFIC_HEAT] = specific_heat_inner_layer  # J/kgK (for Aluminium)
        inner_layer_attributes[Layer.CONVECTION_COEFFICIENT] = convection_coefficient_air_inner_layer  # W/m2K
        inner_layer_attributes[Layer.ABSORPTIVITY] = None  # dimensionless
        inner_layer_attributes[Layer.TEMPERATURE] = temperature  # Initialized with ambient temperature
        inner_layer = Layer(inner_layer_attributes)

        # Layer 2 attributes (mid layer - Rock Wool)
        mid_layer_attributes = dict()
        mid_layer_attributes[Layer.LENGTH] = (outer_length + inner_length) / 2  # m
        mid_layer_attributes[Layer.BREADTH] = (outer_breadth + inner_breadth) / 2  # m
        mid_layer_attributes[Layer.HEIGHT] = (outer_height + inner_height) / 2  # m
        mid_layer_attributes[Layer.THICKNESS] = thickness_mid_layer  # m
        mid_layer_attributes[Layer.DENSITY] = density_mid_layer_material  # kg/m3 (for Rock Wool)
        mid_layer_attributes[Layer.THERMAL_CONDUCTIVITY] = thermal_conductivity_mid_layer  # W/mK  (for Rock Wool)
        mid_layer_attributes[Layer.SPECIFIC_HEAT] = specific_heat_mid_layer  # J/kgK (for Rock Wool)
        mid_layer_attributes[Layer.CONVECTION_COEFFICIENT] = None  # W/m2K
        mid_layer_attributes[Layer.ABSORPTIVITY] = None  # dimensionless
        mid_layer_attributes[Layer.TEMPERATURE] = temperature  # Initialized with ambient temperature
        mid_layer = Layer(mid_layer_attributes)

        # Layer 3 attributes (outer layer - steel)
        outer_layer_attributes = dict()
        outer_layer_attributes[Layer.LENGTH] = outer_length  # m
        outer_layer_attributes[Layer.BREADTH] = outer_breadth  # m
        outer_layer_attributes[Layer.HEIGHT] = outer_height  # m
        outer_layer_attributes[Layer.THICKNESS] = thickness_outer_layer  # m
        outer_layer_attributes[Layer.DENSITY] = density_outer_layer_material  # kg/m3 (for Steel)
        outer_layer_attributes[Layer.THERMAL_CONDUCTIVITY] = thermal_conductivity_outer_layer  # W/mK  (for Steel)
        outer_layer_attributes[Layer.SPECIFIC_HEAT] = specific_heat_outer_layer  # J/kgK (for Steel)
        outer_layer_attributes[Layer.CONVECTION_COEFFICIENT] = convection_coefficient_air_outer_layer  # W/m2K
        outer_layer_attributes[Layer.ABSORPTIVITY] = outer_layer_absorptivity  # dimensionless
        outer_layer_attributes[Layer.TEMPERATURE] = temperature  # Initialized with ambient temperature
        outer_layer = Layer(outer_layer_attributes)

        super().__init__(inner_layer, mid_layer, outer_layer, unavailable_volume, default_scale=1)

    @property
    def azimuth(self) -> float:
        return self.__azimuth

    @property
    def albedo(self) -> float:
        return self.__albedo




from abc import ABC, abstractmethod

from simses.commons.state.system import SystemState


class Auxiliary(ABC):

    def __init__(self):
        pass

    def update(self, time: float, state: SystemState) -> None:
        """
        Updating auxiliary losses of state

        Parameters
        ----------
        state :

        Returns
        -------

        """
        state.aux_losses += self.ac_operation_losses()

    @abstractmethod
    def ac_operation_losses(self) -> float:
        """
        Calculates the ac operation losses

        Parameters
        ----------

        Returns
        -------

        """
        pass

    def close(self) -> None:
        """
        Closing all open resources in operation losses

        Parameters
        ----------

        Returns
        -------

        """
        pass



import numpy as np

from simses.analysis.data.energy_management import EnergyManagementData
from simses.analysis.data.system import SystemData
from simses.analysis.evaluation.economic.revenue_stream.abstract_revenue_stream import RevenueStream
from simses.analysis.evaluation.result import EvaluationResult, Unit, Description
from simses.commons.config.analysis.economic import EconomicAnalysisConfig
from simses.commons.config.analysis.market import MarketProfileConfig
from simses.commons.config.simulation.general import GeneralSimulationConfig
from simses.commons.log import Logger
from simses.commons.profile.economic.constant import ConstantPrice
from simses.commons.profile.economic.fcr import FcrMarketProfile
from simses.commons.profile.economic.market import MarketProfile


class FCRRevenue(RevenueStream):

    def __init__(self, energy_management_data: EnergyManagementData, system_data: SystemData,
                 economic_analysis_config: EconomicAnalysisConfig, general_config: GeneralSimulationConfig,
                 market_profile_config: MarketProfileConfig):
        super().__init__(energy_management_data, system_data, economic_analysis_config)
        self.__log: Logger = Logger(type(self).__name__)
        if economic_analysis_config.fcr_use_price_timeseries:
            self.__fcr_price_profile: MarketProfile = FcrMarketProfile(general_config, market_profile_config)
        else:
            self.__fcr_price_profile: MarketProfile = ConstantPrice(economic_analysis_config.fcr_price)
        self._energy_management_data: EnergyManagementData = energy_management_data
        self.__system_data: SystemData = system_data
        self.__fcr_power_avg = np.array([])
        self.__fcr_price_avg = np.array([])
        self.__fcr_revenue = np.array([])
        self.day_to_sec = 60 * 60 * 24
        self.year_to_sec = self.day_to_sec * 365

        self.__simulation_start = general_config.start
        self.__simulation_end = general_config.end

    def get_cashflow(self) -> np.ndarray:
        time_series = self._energy_management_data.time
        fcr_power_series = abs(self._energy_management_data.fcr_max_power)
        fcr_price_series = []
        cashflow_time_series = []
        fcr_price_scaling_factor_day_to_second = 1 / 1e3 * 1 / self.day_to_sec
        delta_ts = time_series[1] - time_series[0]
        start = self.__simulation_start
        end = self.__simulation_end
        loop = 0
        t_loop = 0

        for time, fcr_power in zip(time_series, fcr_power_series):
            # calculate adapted time for looped simulations (otherwise price file may be out of bounds)
            if (time - start) // (end - start) > loop:
                self.__fcr_price_profile.initialize_profile()
                loop += 1
                t_loop = time
            adapted_time = time - (t_loop - time_series[0])
            price = self.__fcr_price_profile.next(adapted_time)
            fcr_price_series.append(price)
            cashflow_time_series.append(delta_ts * fcr_power * price * fcr_price_scaling_factor_day_to_second)

        self.__fcr_power_avg = np.array(
            self.cash_time_series_to_project_years(fcr_power_series, time_series)) * delta_ts / self.year_to_sec / 1000.0  # in kW
        self.__fcr_price_avg = np.array(self.cash_time_series_to_project_years(fcr_price_series, time_series)) * delta_ts / self.year_to_sec
        self.__fcr_revenue = self.cash_time_series_to_project_years(cashflow_time_series, time_series)
        return np.array(self.__fcr_revenue)

    def get_evaluation_results(self) -> [EvaluationResult]:
        key_results: [EvaluationResult] = list()
        key_results.append(EvaluationResult(Description.Economical.FCR.REVENUE_YEARLY, Unit.EURO, self.__fcr_revenue))
        return key_results

    def get_assumptions(self) -> [EvaluationResult]:
        assumptions: [EvaluationResult] = list()
        assumptions.append(EvaluationResult(Description.Economical.FCR.PRICE_AVERAGE, Unit.EURO_PER_KW_DAY, self.__fcr_price_avg))
        assumptions.append(EvaluationResult(Description.Economical.FCR.POWER_BID_AVERAGE, Unit.KILOWATT, self.__fcr_power_avg ))
        return assumptions

    def close(self):
        self.__log.close()



import pandas as pd
import scipy.interpolate

from simses.commons.config.data.battery import BatteryDataConfig
from simses.commons.config.simulation.battery import BatteryConfig
from simses.commons.log import Logger
from simses.commons.state.technology.lithium_ion import LithiumIonState
from simses.technology.lithium_ion.cell.electric.properties import ElectricalCellProperties
from simses.technology.lithium_ion.cell.format.abstract import CellFormat
from simses.technology.lithium_ion.cell.format.prismatic import PrismaticCell
from simses.technology.lithium_ion.cell.thermal.default import DefaultThermalCellProperties
from simses.technology.lithium_ion.cell.thermal.properties import ThermalCellProperties
from simses.technology.lithium_ion.cell.type import CellType


class SodiumIonGreenRock(CellType):
    # Anode: NaTi2(PO4)3
    # Cathode: : LiMn2O4
    # Electrolyte: Na2SO4-based
    # Separator:
    # current collector: stainless steel
    # at 40°C (= 313,15 K)

    __SOC_HEADER = 'SOC'
    __SOC_IDX = 0
    __OCV_IDX = 1
    __RINT_Ch_IDX = 1
    __RINT_Dch_IDX = 2
    #__SOC_CH_IDX = 0
    #__SOC_DCH_IDX = 2
    #__OCV_CH_IDX = 1
    #__OCV_DCH_IDX = 3

    __CELL_VOLTAGE = 1.5  # V
    __CELL_CAPACITY = 28  # Ah
    __MAX_VOLTAGE: float = 1.9  # V
    __MIN_VOLTAGE: float = 0.8  # V
    __MAX_CHARGE_RATE: float = 0.33  # 1/h
    __MAX_DISCHARGE_RATE: float = 0.33  # 1/h
    __SELF_DISCHARGE_RATE: float = 0.0  # unknown

    __HEIGHT: float = 50  # mm
    __WIDTH: float = 100  # mm
    __LENGTH: float = 200  # mm

    __CELL_FORMAT: CellFormat = PrismaticCell(__HEIGHT, __WIDTH, __LENGTH)

    __COULOMB_EFFICIENCY: float = 1.0  # p.u

    __ELECTRICAL_PROPS: ElectricalCellProperties = ElectricalCellProperties(__CELL_VOLTAGE, __CELL_CAPACITY,
                                                                            __MIN_VOLTAGE, __MAX_VOLTAGE,
                                                                            __MAX_CHARGE_RATE, __MAX_DISCHARGE_RATE,
                                                                            __SELF_DISCHARGE_RATE,
                                                                            __COULOMB_EFFICIENCY)

    __THERMAL_PROPS: ThermalCellProperties = DefaultThermalCellProperties()

    def __init__(self, voltage: float, capacity: float, soh: float, battery_config: BatteryConfig,
                 battery_data_config: BatteryDataConfig):
        super().__init__(voltage, capacity, soh, self.__ELECTRICAL_PROPS, self.__THERMAL_PROPS, self.__CELL_FORMAT,
                         battery_config)

        self.__log: Logger = Logger(type(self).__name__)

        rint_file: str = battery_data_config.sodium_ion_rint_green_rock_file
        ocv_file: str = battery_data_config.sodium_ion_ocv_green_rock_file

        internal_resistance = pd.read_csv(rint_file)  # Ohm
        soc_arr = internal_resistance.iloc[:, self.__SOC_IDX]
        rint_mat_dch = internal_resistance.iloc[:, self.__RINT_Dch_IDX]
        rint_mat_ch = internal_resistance.iloc[:, self.__RINT_Ch_IDX]
        self.__rint_dch_interp1d = scipy.interpolate.interp1d(soc_arr, rint_mat_dch, kind='linear')
        self.__rint_ch_interp1d = scipy.interpolate.interp1d(soc_arr, rint_mat_ch, kind='linear')

        #ocv = pd.read_csv(ocv_file)
        #soc_arr = ocv.iloc[:, self.__SOC_IDX]
        #ocv_mat_ch = ocv.iloc[:, self.__OCV_CH_IDX]
        #ocv_mat_dch = ocv.iloc[:, self.__OCV_IDX]
        #self.__ocv_ch_interp1d = scipy.interpolate.interp1d(soc_arr, ocv_mat_ch, kind='linear')
        #self.__ocv_dch_interp1d = scipy.interpolate.interp1d(soc_arr, ocv_mat_dch, kind='linear')

    def get_open_circuit_voltage(self, battery_state: LithiumIonState) -> float:
        '''Parameters build with ocv fitting'''
        a1 = 4.141
        a2 = -2.883
        a3 = -12.27
        a4 = 21.21
        a5 = -13.75
        a6 = 4.506
        a7 = 0.815

        soc = battery_state.soc

        ocv = a7 + soc*(a6 + a5*soc + a4*soc**2 + a3*soc**3 + a2*soc**4 + a1*soc**5)

        return ocv * self.get_serial_scale()

        #ocv = pd.read_csv(ocv_file)
        #soc_arr = ocv.iloc[:, self.__SOC_IDX]
        #ocv_mat_ch = ocv.iloc[:, self.__OCV_IDX]
        #self.__ocv_ch_interp1d = scipy.interpolate.interp1d(soc_arr, ocv_mat_ch, kind='linear')

    #def get_internal_resistance(self, battery_state: LithiumIonState) -> float:
     #   rint = 0.33784
      #  return float(rint) / self.get_parallel_scale() * self.get_serial_scale()

    def get_internal_resistance(self, battery_state: LithiumIonState) -> float:
        if battery_state.is_charge:
            rint = self.__rint_ch_interp1d(battery_state.soc)
        else:
            rint = self.__rint_dch_interp1d(battery_state.soc)
        return float(rint) / self.get_parallel_scale() * self.get_serial_scale()

    #def get_open_circuit_voltage(self, battery_state: LithiumIonState) -> float:
     #   ocv = self.__ocv_ch_interp1d(battery_state.soc)
      #  return float(ocv) / self.get_parallel_scale() * self.get_serial_scale()

    #def get_open_circuit_voltage(self, battery_state: LithiumIonState) -> float:
     #   if battery_state.is_charge:
      #      ocv = self.__ocv_ch_interp1d(battery_state.soc)
       # else:
        #    ocv = self.__ocv_dch_interp1d(battery_state.soc)
        #return float(ocv) / self.get_parallel_scale() * self.get_serial_scale()

    def close(self) -> None:
        self.__log.close()



from abc import ABC, abstractmethod

from simses.commons.state.technology.fuel_cell import FuelCellState


class FuelCellStackModel(ABC):

    def __init__(self):
        super().__init__()

    def update(self, power: float, state: FuelCellState) -> None:
        """
        Updates current, voltage and hydrogen flow of hydrogen state

        Parameters
        ----------
        power :
        state :

        Returns
        -------

        """
        self.calculate(power)
        state.current = self.get_current()
        state.voltage = self.get_voltage()
        state.hydrogen_use = self.get_hydrogen_consumption()
        # state.power = self.get_power()
        state.current_density = self.get_current_density()


    @abstractmethod
    def calculate(self, power) -> None:
        """
        Calculates current, voltage and hydrogen consumption based on input power
        Parameters
        ----------
        power : input power in W

        Returns
        -------

        """
        pass

    @abstractmethod
    def get_current(self):
        """
        return electrical current in A

        Returns
        -------

        """
        pass

    @abstractmethod
    def get_voltage(self):
        """
        Return of electrical voltage in V

        Returns
        -------

        """
        pass

    @abstractmethod
    def get_hydrogen_consumption(self):
        """
        Return of hydrogen consumption in mol/s

        Returns
        -------

        """
        return self.get_hydrogen_consumption()

    def get_current_density(self) -> float:
        """
        return electrical current of the electrolyzer stack in A cm-2

        Returns
        -------

        """
        return self.get_current() / self.get_geom_area_stack()

    @abstractmethod
    def get_nominal_stack_power(self):
        pass

    @abstractmethod
    def get_number_cells(self):
        pass

    @abstractmethod
    def get_geom_area_stack(self):
        pass


    @abstractmethod
    def get_heat_capactiy_stack(self):
        pass

    @abstractmethod
    def close(self):
        pass



from simses.commons.profile.power.power_profile import PowerProfile
from simses.commons.config.simulation.energy_management import EnergyManagementConfig
from simses.commons.state.energy_management import EnergyManagementState
from simses.commons.state.system import SystemState
from simses.logic.energy_management.strategy.operation_priority import OperationPriority
from simses.logic.energy_management.strategy.operation_strategy import OperationStrategy


class EvChargerWithBuffer(OperationStrategy):
    """
    EvChargerWithBuffer is a basic operation strategy which simulates an EV charging station with a buffer storage.
    The algorithm requires the following profiles and parameters:
    - A load profile that contains the required power for charging the EV
    - The maximal grid power is set in the 'ENERGY_MANAGEMENT' section as 'MAX_POWER'

    Whenever the requested EV power is above the power threshold (depends on data) the EV is charged with the required
    power (Max AC power from grid + buffer storage discharge).
    Whenever the SOC is below the SOC threshold and the power below its threshold the buffer storage is charged with
    the maximal AC power from the grid.
    """

    def __init__(self, power_profile: PowerProfile, ems_config: EnergyManagementConfig):
        super().__init__(OperationPriority.MEDIUM)
        self.__load_profile: PowerProfile = power_profile
        self.__max_ac_grid_power:  float = ems_config.max_power # e.g. 32000
        self.__power: float = 0.0
        self.__soc_threshold: float = 0.98
        self.__power_threshold: float = 50.0

    def next(self, time: float, system_state: SystemState, power: float = 0) -> float:
        self.__power = self.__load_profile.next(time)
        if self.__power > self.__power_threshold or system_state.soc < self.__soc_threshold:
            return self.__max_ac_grid_power - self.__power
        else:
            return 0

    def update(self, energy_management_state: EnergyManagementState) -> None:
        energy_management_state.load_power = self.__power

    def clear(self) -> None:
        self.__power = 0.0

    def close(self) -> None:
        self.__load_profile.close()



from configparser import ConfigParser

from simses.commons.config.data.auxiliary import AuxiliaryDataConfig
from simses.commons.config.data.redox_flow import RedoxFlowDataConfig
from simses.commons.config.simulation.general import GeneralSimulationConfig
from simses.commons.config.simulation.redox_flow import RedoxFlowConfig
from simses.commons.log import Logger
from simses.commons.state.technology.redox_flow import RedoxFlowState
from simses.technology.redox_flow.pump_algorithm.constant_pressure_pulsed import ConstantPressurePulsed
from simses.technology.redox_flow.pump_algorithm.fix_flow_rate_start_stop import FixFlowRateStartStop
from simses.technology.redox_flow.pump_algorithm.abstract_pump_algorithm import PumpAlgorithm
from simses.technology.redox_flow.pump_algorithm.stoich_flow_rate import StoichFlowRate
from simses.system.auxiliary.pump.fixeta_centrifugal import FixEtaCentrifugalPump
from simses.system.auxiliary.pump.abstract_pump import Pump
from simses.system.auxiliary.pump.scalable_variable_eta_centrifugal import ScalableVariableEtaCentrifugalPump
from simses.technology.redox_flow.degradation.const_hydrogen_current import ConstHydrogenCurrent
from simses.technology.redox_flow.degradation.no_degradation import NoDegradation
from simses.technology.redox_flow.degradation.variable_hydrogen_current import VariableHydrogenCurrent
from simses.technology.redox_flow.electrochemical.control.redox_control_system import RedoxControlSystem
from simses.technology.redox_flow.electrochemical.abstract_electrochemical import ElectrochemicalModel
from simses.technology.redox_flow.electrochemical.rint_model import RintModel
from simses.technology.redox_flow.stack.cell_data_stack_5500w import CellDataStack5500W
from simses.technology.redox_flow.stack.dummy_stack_3000w import DummyStack3000W
from simses.technology.redox_flow.stack.dummy_stack_5500W import DummyStack5500W
from simses.technology.redox_flow.stack.electrolyte.abstract_electrolyte import ElectrolyteSystem
from simses.technology.redox_flow.stack.electrolyte.vanadium import VanadiumSystem
from simses.technology.redox_flow.stack.high_performance_stack_5700W import HighPerformanceStack5700W
from simses.technology.redox_flow.stack.industrial_stack_1500w import IndustrialStack1500W
from simses.technology.redox_flow.stack.industrial_stack_9000w import IndustrialStack9000W
from simses.technology.redox_flow.stack.abstract_stack import StackModule


class RedoxFlowFactory:

    def __init__(self, config: ConfigParser):
        self.__log: Logger = Logger(type(self).__name__)
        self.__config_general: GeneralSimulationConfig = GeneralSimulationConfig(config)
        self.__config_redox_flow: RedoxFlowConfig = RedoxFlowConfig(config)
        self.__config_redox_flow_data: RedoxFlowDataConfig = RedoxFlowDataConfig()
        self.__config_auxiliary_data: AuxiliaryDataConfig = AuxiliaryDataConfig()

    def create_redox_flow_state_from(self, storage_id: int, system_id: int, stack_module: StackModule,
                                     capacity: float, temperature: float, redox_flow_state: RedoxFlowState = None):
        """
        Initial creates the RedoxFlowState object if it doesn't exist.

        Parameters
        ----------
        storage_id : int
            storage id
        system_id : int
            system id
        stack_module : StackModule
            stack module based on specific stack typ
        capacity : float
            Capacity of the electrolyte of the redox flow battery in Wh.
        temperature : float
            Temperature of the electrolyte of the redox flow battery in K.
        redox_flow_state : RedoxFlowState

        Returns
        -------
        RedoxFlowState
            state of the redox flow battery
        """
        if redox_flow_state is None:
            time: float = self.__config_general.start
            soc: float = self.__config_redox_flow.soc
            rfbs = RedoxFlowState(system_id, storage_id)
            rfbs.time = time
            rfbs.soc = soc
            rfbs.soc_stack = soc
            rfbs.soh = 1.0
            rfbs.voltage = stack_module.get_open_circuit_voltage(rfbs)
            rfbs.open_circuit_voltage = rfbs.voltage
            rfbs.current = 0.0
            rfbs.capacity = capacity
            rfbs.internal_resistance = stack_module.get_internal_resistance(rfbs)
            rfbs.power_loss = 0.0
            rfbs.power_in = 0.0
            rfbs.power = 0.0
            rfbs.pump_power = 0.0
            rfbs.pressure_loss_anolyte = 0.0
            rfbs.pressure_loss_catholyte = 0.0
            rfbs.pressure_drop_anolyte = 0.0
            rfbs.pressure_drop_catholyte = 0.0
            rfbs.flow_rate_anolyte = 0.0
            rfbs.flow_rate_catholyte = 0.0
            rfbs.fulfillment = 1.0
            rfbs.temperature = temperature
            rfbs.time_pump = 0.0
            rfbs.max_discharge_power = stack_module.get_max_power()
            rfbs.max_charge_power = stack_module.get_max_power()
            return rfbs
        else:
            return redox_flow_state

    def create_stack_module(self, stack_module: str, electrolyte_system: ElectrolyteSystem,
                            voltage: float, power: float) -> StackModule:
        """
        Initial creates the StackModule object for a specific stack typ.

        Parameters
        ----------
        stack_module : str
            stack type for stack module
        electrolyte_system : ElectrolyteSystem
            electrolyte system
        voltage : float
            nominal stack module voltage in V of the redox flow battery
        power : float
            nominal stack module power in W of the redox flow battery

        Returns
        -------
        StackModule
        """
        if stack_module == CellDataStack5500W.__name__:
            self.__log.debug('Creating stack module as ' + stack_module)
            return CellDataStack5500W(electrolyte_system, voltage, power, self.__config_redox_flow_data,
                                      self.__config_redox_flow)
        elif stack_module == DummyStack3000W.__name__:
            self.__log.debug('Creating stack module as ' + stack_module)
            return DummyStack3000W(electrolyte_system, voltage, power, self.__config_redox_flow)
        elif stack_module == DummyStack5500W.__name__:
            self.__log.debug('Creating stack module as ' + stack_module)
            return DummyStack5500W(electrolyte_system, voltage, power, self.__config_redox_flow)
        elif stack_module == IndustrialStack1500W.__name__:
            self.__log.debug('Creating stack module as ' + stack_module)
            return IndustrialStack1500W(electrolyte_system, voltage, power, self.__config_redox_flow_data,
                                        self.__config_redox_flow)
        elif stack_module == IndustrialStack9000W.__name__:
            self.__log.debug('Creating stack module as ' + stack_module)
            return IndustrialStack9000W(electrolyte_system, voltage, power, self.__config_redox_flow_data,
                                        self.__config_redox_flow)
        elif stack_module == HighPerformanceStack5700W.__name__:
            self.__log.debug('Creating stack module as ' + stack_module)
            return HighPerformanceStack5700W(electrolyte_system, voltage, power, self.__config_redox_flow_data,
                                             self.__config_redox_flow)
        else:
            options: [str] = list()
            options.append(CellDataStack5500W.__name__)
            options.append(DummyStack3000W.__name__)
            options.append(DummyStack5500W.__name__)
            options.append(IndustrialStack1500W.__name__)
            options.append(IndustrialStack9000W.__name__)
            options.append(HighPerformanceStack5700W.__name__)
            raise Exception('Specified stack module ' + stack_module + ' is unknown. '
                            'Following options are available: ' + str(options))

    def check_pump_algorithm(self, pump_algorithm: str, stack_module) -> str:
        if pump_algorithm == 'Default':
            if stack_module == CellDataStack5500W.__name__ or stack_module == DummyStack3000W.__name__ \
                    or stack_module == DummyStack5500W.__name__ or stack_module == HighPerformanceStack5700W.__name__:
                return StoichFlowRate.__name__
            elif stack_module == IndustrialStack1500W.__name__ or stack_module == IndustrialStack9000W.__name__:
                return ConstantPressurePulsed.__name__
        else:
            return pump_algorithm

    def create_pumps(self, pump_algorithm: str) -> Pump:
        if pump_algorithm == FixFlowRateStartStop.__name__:
            pump_type = FixEtaCentrifugalPump.__name__
            self.__log.debug('Creating pumps as ' + pump_type)
            # Value based on data from König (2017) (Dissertation)
            efficiency: float = 0.345
            return FixEtaCentrifugalPump(efficiency)
        # select if a special pump with a data file should be used
        # elif pump_algorithm == ConstantPressurePulsed.__name__:
        #     pump_type = 'VariableEtaCentrifugalPump'
        #     self.__log.debug('Creating pumps as ' + pump_type)
        #     return VariableEtaCentrifugalPump(self.__config_auxiliary_data)
        elif pump_algorithm == StoichFlowRate.__name__ or pump_algorithm == ConstantPressurePulsed.__name__:
            pump_type = ScalableVariableEtaCentrifugalPump.__name__
            self.__log.debug('Creating pumps as ' + pump_type)
            return ScalableVariableEtaCentrifugalPump()
        else:
            options: [str] = list()
            options.append(StoichFlowRate.__name__)
            options.append(FixFlowRateStartStop.__name__)
            options.append(ConstantPressurePulsed.__name__)
            raise Exception('Specified pump algorithm ' + pump_algorithm + ' is unknown. '
                            'Following options are available: ' + str(options))

    def create_electrolyte_system(self, capacity: float, stack_type: str):
        """
        Initial creates the ElectrolyteSystem object.

        Parameters
        ----------
        capacity : float
            Total start capacity of the redox flow system in Wh.
        stack_type: str
            stack type of the redox flow battery

        Returns
        -------
        ElectrolyteSystem
        """
        if (stack_type == CellDataStack5500W.__name__ or stack_type == DummyStack3000W.__name__ or
            stack_type == DummyStack5500W.__name__ or stack_type == IndustrialStack1500W.__name__ or
            stack_type == IndustrialStack9000W.__name__ or stack_type == HighPerformanceStack5700W.__name__):
            self.__log.debug('Creating electrolyte system as VanadiumSystem')
            return VanadiumSystem(capacity, self.__config_redox_flow)
        else:
            options: [str] = list()
            options.append(CellDataStack5500W.__name__)
            options.append(DummyStack3000W.__name__)
            options.append(DummyStack5500W.__name__)
            options.append(IndustrialStack1500W.__name__)
            options.append(IndustrialStack9000W.__name__)
            options.append(HighPerformanceStack5700W.__name__)
            raise Exception('Specified stack module for vanadium system ' + stack_type + ' is unknown. '
                            'Following options are available: ' + str(options))

    def create_electrochemical_model(self, stack_module: StackModule,
                                     battery_management_system: RedoxControlSystem
                                     , electrolyte_system: ElectrolyteSystem, pump_algorithm: PumpAlgorithm,
                                     electrochemical_model: ElectrochemicalModel = None) -> ElectrochemicalModel:
        """
        Initial creates the ElectrochemicalModel object for a specific model, which includes the battery management
        system requests.

        Parameters
        ----------
        stack_module : StackModule
            stack module of a redox flow battery
        battery_management_system : RedoxControlSystem
            battery management system of the redox flow battery
        electrolyte_system: ElectrolyteSystem
            electrolyte system of the redox flow battery
        pump_algorithm: PumpAlgorithm
            pump algorithm ot the redox flow battery
        electrochemical_model : ElectrochemicalModel
            electrochemical model of the redox flow battery

        Returns
        -------
            ElectrochemicalModel
        """
        if electrochemical_model is None:
            self.__log.debug('Creating electrochemical model for redox flow system depended on stack module ' +
                             stack_module.__class__.__name__)
            return RintModel(stack_module, battery_management_system, electrolyte_system, pump_algorithm)
        else:
            return electrochemical_model

    def create_redox_management_system(self, stack_module: StackModule, electrolyte_system: ElectrolyteSystem,
                                       pump_algorithm: PumpAlgorithm,
                                       redox_management_system: RedoxControlSystem = None) \
            -> RedoxControlSystem:
        """
        Initial creates the BatteryManagementSystem object of the redox flow battery.

        Parameters
        ----------
        stack_module : StackModule
            stack module of a redox flow battery
        electrolyte_system : ElectrolyteSystem
            management system of a redox flow battery
        pump_algorithm : PumpAlgorithm
            pump algorithm of the redox flow battery
        redox_management_system : RedoxControlSystem
            battery management system of the redox flow battery

        Returns
        -------
            RedoxControlSystem
        """
        if redox_management_system is None:
            self.__log.debug('Creating battery management system for redox flow system depended on stack module '
                             + stack_module.__class__.__name__)
            return RedoxControlSystem(stack_module, electrolyte_system, pump_algorithm, self.__config_redox_flow)
        else:
            return redox_management_system

    def create_degradation_model(self, capacity, stack_module: StackModule):
        # Degradation model can be changed here.
        degradation_model = 'ConstHydrogenCurrent'
        if degradation_model == 'NoDegradation':
            self.__log.debug('Creating degradation Model as ' + degradation_model)
            return NoDegradation(capacity)
        elif degradation_model == 'ConstHydrogenCurrent':
            self.__log.debug('Creating degradation Model as ' + degradation_model)
            return ConstHydrogenCurrent(capacity, stack_module)
        elif degradation_model == 'VariableHydrogenCurrent':
            self.__log.debug('Creating degradation Model as ' + degradation_model)
            return VariableHydrogenCurrent(capacity, stack_module, self.__config_redox_flow_data)
        else:
            options: [str] = list()
            options.append(NoDegradation.__name__)
            options.append(ConstHydrogenCurrent.__name__)
            options.append(VariableHydrogenCurrent.__name__)
            raise Exception('Specified degradation model ' + degradation_model + ' is unknown. '
                            'Following options are available: ' + str(options))

    def create_pump_algorithm(self, pump: Pump, stack_module: StackModule, electrolyte_system: ElectrolyteSystem,
                              pump_algorithm: str):
        if pump_algorithm == StoichFlowRate.__name__:
            self.__log.debug('Creating pump algorithm as ' + pump_algorithm)
            return StoichFlowRate(stack_module, pump, electrolyte_system, self.__config_redox_flow)
        elif pump_algorithm == FixFlowRateStartStop.__name__:
            self.__log.debug('Creating pump algorithm as ' + pump_algorithm)
            return FixFlowRateStartStop(stack_module, pump, electrolyte_system)
        elif pump_algorithm == ConstantPressurePulsed.__name__:
            self.__log.debug('Creating pump algorithm as ' + pump_algorithm)
            time_step = self.__config_general.timestep
            return ConstantPressurePulsed(stack_module, pump, electrolyte_system, int(time_step))
        else:
            options: [str] = list()
            options.append(StoichFlowRate.__name__)
            options.append(FixFlowRateStartStop.__name__)
            options.append(ConstantPressurePulsed.__name__)
            raise Exception('Specified pump algorithm ' + pump_algorithm + ' is unknown. '
                            'Following options are available: ' + str(options))

    def close(self):
        self.__log.close()



from simses.commons.state.technology.storage import StorageTechnologyState
from .cycle_detector import CycleDetector


class NoCycleDetector(CycleDetector):

    def __init__(self):
        super().__init__()

    def cycle_detected(self, time: float, state: StorageTechnologyState) -> bool:
        return False

    def get_depth_of_cycle(self) -> float:
        return 0

    def get_delta_full_equivalent_cycle(self) -> float:
        return 0

    def get_crate(self) -> float:
        return 0

    def get_full_equivalent_cycle(self) -> float:
        return 0

    def get_mean_soc(self) -> float:
        return 0

    def reset(self) -> None:
        pass



from simses.system.auxiliary.heating_ventilation_air_conditioning.hvac import HeatingVentilationAirConditioning


class NoHeatingVentilationAirConditioning(HeatingVentilationAirConditioning):

    def __init__(self):
        super().__init__()

    def run_air_conditioning(self, temperature_time_series: [float], temperature_timestep: float, ambient_air_temperature: float) -> float:
        pass

    def get_electric_power(self) -> float:
        return 0

    def get_max_thermal_power(self) -> float:
        return 0

    def get_set_point_temperature(self) -> float:
        return 298.15

    def get_thermal_power(self) -> float:
        return 0

    def get_cop(self) -> float:
        return 1.0

    def get_scop(self) -> float:
        return 1.0

    def get_seer(self) -> float:
        return 1.0

    def get_temperature_dead_band(self) -> float:
        return 0.0

    def update_air_parameters(self, air_mass: float = None, air_specific_heat: float = None, air_density: float = None) -> None:
        pass

    def set_electric_power(self, electric_power: float) -> None:
        """Used in cases where multiple runs of the HVAC take place within a SimSES timestep"""
        pass



from simses.technology.hydrogen.control.pressure.pressure_controller import \
    PressureController


class NoPressureController(PressureController):

    def __init__(self):
        super().__init__()

    def calculate_n_h2_out(self, pressure_cathode: float, n_h2_prod: float, timestep: float, pressure_factor: float) -> float:
        return n_h2_prod

    def calculate_n_o2_out(self, pressure_anode: float, n_o2_prod: float, timestep: float) -> float:
        return n_o2_prod

    def calculate_n_h2_in(self, pressure_anode: float, n_h2_used: float) -> float:
        return n_h2_used

    def calculate_n_o2_in(self, pressure_cathode: float, n_o2_used: float) -> float:
        return n_o2_used



from simses.commons.log import Logger
from simses.system.auxiliary.pump.abstract_pump import Pump


class ScalableVariableEtaCentrifugalPump(Pump):
    """ScalableVariableEtaCentrifugalPump is a pump with an efficiency dependent on the flow rate. The pump is scaled
    based on the maximal and minimal expected flow rate in the system."""

    __ALPHA = 1.082
    __BETA = 1.223

    def __init__(self):
        """
        The efficiency curve of the pump was generated from data from König (2017).
        Literature source: König, Sebastian. "Model-based design and optimization of vanadium redox flow batteries."
        (2017) Karlsruhe Institute of Technology. Dissertation"""
        super().__init__()
        self.__log: Logger = Logger(type(self).__name__)
        self.__power = 0
        self.__eta = 0

    def calculate_pump_power(self, pressure_loss) -> None:
        if pressure_loss == 0:
            self.__power = 0
        else:
            self.__power = pressure_loss / self.__eta

    def set_eta_pump(self, flow_rate, flow_rate_max, flow_rate_min) -> None:
        flow_rate_ratio_max_eff = self.__BETA/(2*self.__ALPHA)
        if (flow_rate_min + flow_rate_max)/(2*flow_rate_max) < flow_rate_ratio_max_eff:
            flow_rate_max_pump = flow_rate_max
        else:
            flow_rate_max_pump = (flow_rate_min + flow_rate_max)/(2*flow_rate_ratio_max_eff)
        flow_rate_ratio = flow_rate/flow_rate_max_pump
        if flow_rate > flow_rate_max_pump:
            self.__log.error('Maximal flow rate for pump is to low. Check maximal flow rate calculation of the system.')
        self.__eta = -self.__ALPHA * flow_rate_ratio ** 2 + self.__BETA * flow_rate_ratio
        if self.__eta < 0 or self.__eta > 1:
            self.__log.error('Pump efficiency has to be greater than 0 and smaller or equal to 1. It is set to 1.')
            self.__eta = 1
        self.__log.debug('Flow rate is ' + str(flow_rate) + '. Pump efficiency is ' + str(self.__eta))

    def get_pump_power(self) -> float:
        return self.__power

    def close(self):
        super().close()
        self.__log.close()



import pandas as pd
import scipy.interpolate

from simses.commons.log import Logger
from simses.system.auxiliary.pump.abstract_pump import Pump
from simses.commons.config.data.auxiliary import AuxiliaryDataConfig


class VariableEtaCentrifugalPump(Pump):
    """VariableEtaCentrifugalPump is a pump with an efficiency dependent on the flow rate."""

    def __init__(self, auxiliary_data_config: AuxiliaryDataConfig):
        super().__init__()
        self.__log: Logger = Logger(type(self).__name__)
        self.__power = 0
        self.__eta = 0
        self.__ETA_FILE = auxiliary_data_config.pump_eta_file
        self.__efficiency = pd.read_csv(self.__ETA_FILE, header=None, index_col=0)
        self.__flow_rate_arr = self.__efficiency.iloc[0] * 10 ** -3 / 60  # m^3/s
        self.__eta_arr = self.__efficiency.iloc[1]/100
        self.__efficiency_interp1d = scipy.interpolate.interp1d(self.__flow_rate_arr, self.__eta_arr, kind='linear')
        self.__max_pump_flow_rate = max(self.__flow_rate_arr)
        self.__min_pump_flow_rate = min(self.__flow_rate_arr)

    def calculate_pump_power(self, pressure_loss) -> None:
        if pressure_loss < 0:
            self.__log.error('Pressure losses are negative. Check sign of pressure drop and flow rate.')
        self.__power = pressure_loss / self.__eta

    def set_eta_pump(self, flow_rate, flow_rate_max, flow_rate_min) -> None:
        self.__log.debug('Flow rate is ' + str(flow_rate))
        if self.__min_pump_flow_rate > flow_rate > self.__max_pump_flow_rate:
            self.__log.error('Flow rate is to high or low for the pump. Use different pump data. '
                             'Pump efficiency is set to 1.')
            self.__eta = 1
        elif flow_rate == 0:
            self.__eta = 1
        else:
            self.__eta = self.__efficiency_interp1d(flow_rate)

        if self.__eta < 0 or self.__eta > 1:
            self.__log.error('Pump efficiency has to be greater than 0 and smaller or equal to 1. It is set to 1.')
            self.__eta = 1
        self.__log.debug('Pump Efficiency is ' + str(self.__eta))

    def get_pump_power(self) -> float:
        return self.__power

    def close(self):
        super().close()
        self.__log.close()



from simses.commons.state.energy_management import EnergyManagementState
from simses.commons.state.system import SystemState
from simses.logic.energy_management.strategy.operation_strategy import OperationStrategy


class StackedOperationStrategy(OperationStrategy):
    """
    Algorithm to determine output power for each operation (in multiuse case) of the ESS for the next timestep.
    """

    def __init__(self, priority: float, strategies: []):
        super().__init__(priority)
        OperationStrategy.sort(strategies)
        self.__strategies: [] = strategies

    def next(self, time: float, system_state: SystemState, power: float = 0) -> float:
        for strategy in self.__strategies:
            power += strategy.next(time, system_state, power)
        return power

    def update(self, energy_management_state: EnergyManagementState) -> None:
        for strategy in self.__strategies:
            strategy.update(energy_management_state)

    def clear(self) -> None:
        for strategy in self.__strategies:
            strategy.clear()

    def close(self) -> None:
        for strategy in self.__strategies:
            strategy.close()



from configparser import ConfigParser
from simses.commons.data.data_handler import DataHandler
from simses.commons.error import EndOfLifeError
from simses.commons.log import Logger
from simses.commons.state.parameters import SystemParameters
from simses.commons.state.technology.lithium_ion import LithiumIonState
from simses.system.auxiliary.auxiliary import Auxiliary
from simses.technology.lithium_ion.battery_management_system.management_system import BatteryManagementSystem
from simses.technology.lithium_ion.cell.type import CellType
from simses.technology.lithium_ion.degradation.degradation_model import DegradationModel
from simses.technology.lithium_ion.equivalent_circuit_model.equivalent_circuit import EquivalentCircuitModel
from simses.technology.lithium_ion.factory import LithiumIonFactory
from simses.technology.storage import StorageTechnology

from scipy.optimize import fsolve

class LithiumIonBattery(StorageTechnology):
    """Battery orchestrates its models for lithium_ion management system, degradation and thermal management as well as
    its equivalent circuit"""

    __ACCURACY: float = 1e-14
    __MAX_LOOPS: int = 40
    __SHARE: float = 0.5

    def __init__(self, cell: str, voltage: float, capacity: float, soc: float, soh: float, data_export: DataHandler,
                 temperature: float, storage_id: int, system_id: int, config: ConfigParser):
        super().__init__()
        self.__log: Logger = Logger(type(self).__name__)
        self.__factory: LithiumIonFactory = LithiumIonFactory(config)
        cell_type: CellType = self.__factory.create_cell_type(cell, voltage, capacity, soh)
        self.__cell_type: CellType = cell_type
        self.__battery_state: LithiumIonState = self.__factory.create_battery_state_from(system_id,
                                                                                         storage_id,
                                                                                         cell_type,
                                                                                         temperature,
                                                                                         soc)
        self.__degradation_model: DegradationModel = self.__factory.create_degradation_model_from(cell_type,
                                                                                                  self.__battery_state)
        self.__battery_management_system: BatteryManagementSystem = \
            self.__factory.create_battery_management_system_from(cell_type)
        self.__battery_model: EquivalentCircuitModel = self.__factory.create_battery_model_from(cell_type)
        self.__data_export: DataHandler = data_export
        self.__log.debug('created: ' + str(self.__battery_state))
        self.__export_data(self.__battery_state)  # initial timestep

    def get_equilibrium_state_for(self, time: float, current: float, fixed_values: bool = False) -> LithiumIonState:
        """Starting update of lithium_ion"""
        bs: LithiumIonState = LithiumIonState(0, 0)
        bs.set_all(self.__battery_state)

        # find current that fulfills power target
        bs.current = current
        power_target: float = bs.current * bs.voltage
        if not fixed_values:
            self._equilibrium_fsolve(bs, power_target)

        # check current limits, update voltage and losses
        self.__battery_management_system.update(time, bs, power_target)
        # update soc, ocv and rint
        self.__battery_model.update(time, bs)
        
        return bs

    def _equilibrium_fsolve(self, state: LithiumIonState, p : float) -> None:
        """Find the current that fulfills the power target, equivalent circuit model equilibrium
        
        p = i * v 
          = i * (ocv + r * i) 
        
        """

        ocv = state.voltage_open_circuit
        r = state.internal_resistance

        # fsolve result is 1-element np-array, take first element for result type float
        i = fsolve(lambda i: p - i * (ocv + i * r), x0=0.0)[0] 
        v = ocv + i * r

        state.current = i

    def distribute_and_run(self, time: float, current: float, voltage: float):
        bs: LithiumIonState = self.get_equilibrium_state_for(time, current)
        end_of_life_reached: bool = False
        try:
            self.__degradation_model.update(time, bs)
        except EndOfLifeError:
            end_of_life_reached = True
        finally:
            # TODO update battery temp in system thermal model
            bs.time = time
            self.__export_data(bs)
            self.__battery_state.set_all(bs)
        if end_of_life_reached:
            raise EndOfLifeError()

    def __export_data(self, state: LithiumIonState) -> None:
        self.__data_export.transfer_data(state.to_export())

    @property
    def volume(self) -> float:
        return self.__cell_type.get_volume()

    @property
    def mass(self) -> float:
        return self.__cell_type.get_mass()

    @property
    def surface_area(self) -> float:
        return self.__cell_type.get_surface_area()

    @property
    def specific_heat(self) -> float:
        return self.__cell_type.get_specific_heat()

    @property
    def convection_coefficient(self) -> float:
        return self.__cell_type.get_convection_coefficient()

    @property
    def max_voltage(self) -> float:
        return self.__cell_type.get_max_voltage()

    @property
    def min_voltage(self) -> float:
        return self.__cell_type.get_min_voltage()

    @property
    def max_current(self) -> float:
        return self.__cell_type.get_max_current(self.__battery_state)

    @property
    def min_current(self) -> float:
        return self.__cell_type.get_min_current(self.__battery_state)

    @property
    def id(self) -> int:
        return int(self.__battery_state.get(LithiumIonState.SYSTEM_DC_ID))

    def wait(self):
        pass

    def get_auxiliaries(self) -> [Auxiliary]:
        return list()

    @property
    def state(self) -> LithiumIonState:
        return self.__battery_state

    def get_system_parameters(self) -> dict:
        parameters: dict = dict()
        parameters[SystemParameters.CELL_TYPE] = self.__cell_type.get_name()
        parameters[SystemParameters.NOMINAL_VOLTAGE] = int(self.__cell_type.get_nominal_voltage())
        parameters[SystemParameters.BATTERY_CIRCUIT] = str(self.__cell_type.get_serial_scale()) + 's' + \
                                                       str(self.__cell_type.get_parallel_scale()) + 'p'
        return parameters

    def close(self) -> None:
        """Closing all resources in lithium_ion"""
        self.__battery_management_system.close()
        self.__battery_model.close()
        self.__degradation_model.close()
        self.__factory.close()
        self.__log.close()
